from __future__ import print_function
from __future__ import division

import pickle
import io
import warnings

import numpy as np
import pandas as pd
from numpy.testing import assert_almost_equal
from sklearn.exceptions import NotFittedError
from sklearn.utils.estimator_checks import (
    check_transformer_data_not_an_array,
    check_transformer_general,
    check_transformers_unfitted,
    check_transformer_n_iter,
    check_fit2d_predict1d,
    check_fit2d_1sample,
    check_fit2d_1feature,
    check_fit1d_1feature,
    check_fit1d_1sample,
    check_get_params_invariance,
    check_dict_unchanged,
    check_dont_overwrite_parameters,
    check_parameters_default_constructible,
    check_no_fit_attributes_set_in_init)
import pytest

from civismlext.preprocessing import DataFrameETL


NAN_STRING = 'NaN_Sentinel'


@pytest.fixture()
def data_raw():
    raw = pd.concat([
        pd.Series(['a', 'b', 'c'], dtype='category', name='pid'),
        pd.Series(['marid', 'effrit', 'sila'], dtype='object',
                  name='djinn_type'),
        pd.Series([1.0, np.NaN, 3.0], dtype='float', name='fruits'),
        pd.Series([2000, 2500, 3000], dtype='uint16', name='age'),
        pd.Series(['cat', 'dog', np.NaN], dtype='category', name='animal'),
    ], axis=1)
    return raw


@pytest.fixture()
def dataframe_expected():
    df = pd.concat([
        pd.Series([0., 1., 0.], dtype='float32', name='djinn_type_effrit'),
        pd.Series([1., 0., 0.], dtype='float32', name='djinn_type_marid'),
        pd.Series([0., 0., 1.], dtype='float32', name='djinn_type_sila'),
        pd.Series([0., 0., 0.], dtype='float32', name='djinn_type_NaN'),
        pd.Series([1., 0., 0.], dtype='float32', name='fruits_1.0'),
        pd.Series([0., 0., 1.], dtype='float32', name='fruits_3.0'),
        pd.Series([0., 1., 0.], dtype='float32', name='fruits_NaN'),
        pd.Series([2000., 2500., 3000.], dtype='float32', name='age'),
        pd.Series([1., 0., 0.], dtype='float32', name='animal_cat'),
        pd.Series([0., 1., 0.], dtype='float32', name='animal_dog'),
        pd.Series([0., 0., 1.], dtype='float32', name='animal_NaN'),
    ], axis=1)
    return df


@pytest.fixture()
def data_raw_2():
    raw = pd.concat([
        pd.Series(['c', np.NaN, 'NaN_Sentinel'], dtype='category', name='pid'),
        pd.Series(['marid', 'effrit', 'sila'], dtype='object',
                  name='djinn_type'),
        pd.Series([-99999.0, np.NaN, 1.0], dtype='float', name='fruits'),
        pd.Series([2000, 2500, 3000], dtype='uint16', name='age'),
        pd.Series(['cat', 'dog', np.NaN], dtype='category', name='animal'),
    ], axis=1)
    return raw


@pytest.fixture()
def dataframe_2_expected():
    df = pd.concat([
        pd.Series([0., 0., 0.], dtype='float32', name='pid_a'),
        pd.Series([0., 0., 0.], dtype='float32', name='pid_b'),
        pd.Series([1., 0., 0.], dtype='float32', name='pid_c'),
        pd.Series([0., 1., 0.], dtype='float32', name='pid_NaN'),
        pd.Series([0., 0., 1.], dtype='float32', name='fruits_1.0'),
        pd.Series([0., 0., 0.], dtype='float32', name='fruits_3.0'),
        pd.Series([0., 1., 0.], dtype='float32', name='fruits_NaN')
    ], axis=1)
    return df


@pytest.fixture()
def data_few_levels():
    raw = pd.concat([
        pd.Series(['a', 'NaN', 'NaN'], dtype='category', name='pid'),
        pd.Series([1.0, np.NaN, 1.0], dtype='float', name='fruits'),
        pd.Series(['cat', 'cat', 'cat'], dtype='category', name='animal'),
    ], axis=1)
    return raw


@pytest.fixture()
def few_levels_expected():
    df = pd.concat([
        pd.Series([0., 1., 1.], dtype='float32', name='pid_NaN'),
        pd.Series([1., 0., 0.], dtype='float32', name='pid_a'),
        pd.Series([0., 0., 0.], dtype='float32', name='pid_NaN_Sentinel'),
        pd.Series([1., 99., 1.], dtype='float32', name='fruits_1.0'),
        pd.Series([0., 1., 0.], dtype='float32', name='fruits_NaN'),
        pd.Series([1., 1., 1.], dtype='float32', name='animal_cat'),
        pd.Series([0., 0., 0.], dtype='float32', name='animal_NaN')
    ], axis=1)
    return df


@pytest.fixture()
def levels_dict():
    levels = {
        'pid': ['a', 'b', 'c', NAN_STRING],
        'djinn_type': ['effrit', 'marid', 'sila', NAN_STRING],
        'animal': ['cat', 'dog', NAN_STRING]
    }
    return levels


@pytest.fixture()
def levels_dict_numeric():
    levels = {'pid': ['a', 'b', 'c', NAN_STRING],
              'fruits': [1.0, 3.0, NAN_STRING]}
    return levels


def test_sklearn_api():
    name = DataFrameETL.__name__
    check_parameters_default_constructible(name, DataFrameETL)
    check_no_fit_attributes_set_in_init(name, DataFrameETL)

    estimator = DataFrameETL()

    for check in [
            check_transformers_unfitted,
            check_transformer_n_iter,
            check_get_params_invariance]:
        check(name, estimator)

    # these are known failures
    for check in [
            check_transformer_data_not_an_array,
            check_transformer_general,
            check_fit2d_predict1d,
            check_fit2d_1sample,
            check_fit2d_1feature,
            check_fit1d_1feature,
            check_fit1d_1sample,
            check_dict_unchanged,
            check_dont_overwrite_parameters]:
        with pytest.raises(TypeError) as e:
            check(name, estimator)
        assert (
            "ETL transformer must be fit "
            "to a dataframe.") in str(e.value)


def test_flag_nulls_warn(data_raw):
    expander = DataFrameETL(check_null_cols='warn')
    drop_cols = ['col1', 'col2']
    drop_cols_2 = ['col1', 'col2', 'nantastic']
    assert expander._flag_nulls(data_raw, []) == []
    assert expander._flag_nulls(data_raw, drop_cols) == drop_cols

    # add a col of all nans
    data_raw['nantastic'] = pd.Series([np.NaN] * 3)
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # check that we don't add the col if it's already being dropped
        assert expander._flag_nulls(data_raw, drop_cols_2) == drop_cols_2
        assert len(w) == 0

        assert expander._flag_nulls(data_raw, drop_cols) == drop_cols_2
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)


def test_flag_nulls_raise(data_raw):
    expander = DataFrameETL(check_null_cols='raise')
    drop_cols = ['col1', 'col2']

    # add a col of all nans
    data_raw['nantastic'] = pd.Series([np.NaN] * 3)
    with pytest.raises(RuntimeError):
        expander._flag_nulls(data_raw, drop_cols)


def test_check_sentinels(data_raw):
    expander = DataFrameETL(cols_to_expand=['pid', 'djinn_type',
                                            'animal', 'fruits'])
    # fill in necessary parameters
    expander._nan_sentinel = 'effrit'
    expander.levels_ = {}
    expander._cols_to_drop = expander.cols_to_drop
    expander._cols_to_expand = expander.cols_to_expand
    expander._check_sentinels(data_raw)
    assert expander._nan_sentinel is not 'effrit'
    assert not (data_raw[['pid', 'djinn_type', 'animal']] ==
                expander._nan_sentinel).any().any()

    # The NaN sentinel can't be in the "fruits" column because
    # "fruits" is numeric and the sentinel is not.
    assert np.issubdtype(data_raw['fruits'].dtype, np.number)
    assert not np.issubdtype(type(expander._nan_sentinel), np.number)


def test_create_levels(data_raw, levels_dict):
    expander = DataFrameETL(cols_to_expand=['pid', 'djinn_type', 'animal'])
    expander._nan_sentinel = NAN_STRING
    expander._cols_to_drop = expander.cols_to_drop
    expander._cols_to_expand = expander.cols_to_expand
    actual_levels = expander._create_levels(data_raw)
    assert actual_levels == levels_dict


def test_create_levels_no_dummy(data_raw, levels_dict_numeric):
    expander = DataFrameETL(cols_to_expand=['pid', 'fruits'],
                            dummy_na=False)
    # remove nan from pid levels
    levels_dict_numeric['pid'] = ['a', 'b', 'c']
    expander._nan_sentinel = NAN_STRING
    expander._cols_to_drop = expander.cols_to_drop
    expander._cols_to_expand = expander.cols_to_expand
    actual_levels = expander._create_levels(data_raw)
    assert actual_levels == levels_dict_numeric


def test_warn_too_many_categories():
    df = pd.DataFrame({'cat': list(range(2000)),
                       'bird': 2 * list(range(1000))})
    with pytest.warns(RuntimeWarning) as warn:
        DataFrameETL(cols_to_expand=['cat', 'bird']).fit(df)
    assert len(warn) == 1, "Should only raise one warning"
    assert '"cat": 2000 categories' in warn[0].message.args[0]
    assert '"bird": 1000 categories' in warn[0].message.args[0]


def test_create_col_names(data_raw):
    expander = DataFrameETL(cols_to_expand=['pid', 'djinn_type', 'animal'],
                            cols_to_drop=['fruits'],
                            dummy_na=True)
    expander._nan_sentinel = NAN_STRING
    expander._cols_to_drop = expander.cols_to_drop
    expander._cols_to_expand = expander.cols_to_expand
    expander.levels_ = expander._create_levels(data_raw)
    (cnames, unexpanded) = expander._create_col_names(data_raw)
    cols_expected = ['pid_a', 'pid_b', 'pid_c', 'pid_NaN',
                     'djinn_type_effrit', 'djinn_type_marid',
                     'djinn_type_sila', 'djinn_type_NaN', 'age',
                     'animal_cat', 'animal_dog', 'animal_NaN']
    assert cnames == cols_expected
    assert unexpanded == ['pid', 'djinn_type', 'age', 'animal']


def test_create_col_names_no_dummy(data_raw):
    expander = DataFrameETL(cols_to_expand=['pid', 'djinn_type', 'animal'],
                            cols_to_drop=['fruits'],
                            dummy_na=False)
    expander._nan_sentinel = NAN_STRING
    expander._cols_to_drop = expander.cols_to_drop
    expander._cols_to_expand = expander.cols_to_expand
    expander.levels_ = expander._create_levels(data_raw)
    (cnames, unexpanded) = expander._create_col_names(data_raw)
    cols_expected = ['pid_a', 'pid_b', 'pid_c',
                     'djinn_type_effrit', 'djinn_type_marid',
                     'djinn_type_sila', 'age',
                     'animal_cat', 'animal_dog']
    assert cnames == cols_expected
    assert unexpanded == ['pid', 'djinn_type', 'age', 'animal']


def test_create_col_names_numeric(data_raw):
    expander = DataFrameETL(cols_to_expand=['pid', 'fruits'],
                            cols_to_drop=['djinn_type', 'animal'],
                            dummy_na=True)
    expander._nan_sentinel = NAN_STRING
    expander._cols_to_drop = expander.cols_to_drop
    expander._cols_to_expand = expander.cols_to_expand
    expander.levels_ = expander._create_levels(data_raw)
    (cnames, unexpanded) = expander._create_col_names(data_raw)
    cols_numeric = ['pid_a', 'pid_b', 'pid_c', 'pid_NaN', 'fruits_1.0',
                    'fruits_3.0', 'fruits_NaN', 'age']
    assert cnames == cols_numeric
    assert unexpanded == ['pid', 'fruits', 'age']


def test_dropped_cols_no_levels(data_raw):
    # If the user requests that we drop a column, we shouldn't create
    # levels for it. That risks raising a warning for too many levels
    # when it doesn't matter.
    expander = DataFrameETL(cols_to_drop=['pid'])
    expander.fit(data_raw)

    assert 'animal' in expander.levels_
    assert 'pid' not in expander.levels_


def test_expand_col(data_raw):
    expander = DataFrameETL(cols_to_drop=['fruits'],
                            dummy_na=True,
                            fill_value=-1.0)
    expander.fit(data_raw)
    # should expand even if there are no NaNs
    arr = expander._expand_col(data_raw, 'pid')
    arr_exp = np.array([[1., 0., 0., 0.], [0., 1., 0., 0.],
                        [0., 0., 1., 0.]])
    assert_almost_equal(arr, arr_exp)
    arr = expander._expand_col(data_raw, 'animal')
    arr_exp = np.array([[1., 0., 0.], [0., 1., 0.], [-1., -1., 1.]])
    assert_almost_equal(arr, arr_exp)


def test_expand_col_no_dummy(data_raw):
    expander = DataFrameETL(cols_to_drop=['fruits'],
                            dummy_na=False,
                            fill_value=-1.0)
    expander.fit(data_raw)
    arr = expander._expand_col(data_raw, 'pid')
    arr_exp = np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
    assert_almost_equal(arr, arr_exp)
    arr = expander._expand_col(data_raw, 'animal')
    arr_exp = np.array([[1., 0.], [0., 1.], [-1., -1.]])
    assert_almost_equal(arr, arr_exp)


def test_expand_col_numeric(data_raw):
    expander = DataFrameETL(cols_to_drop=['pid', 'djinn_type', 'animal'],
                            cols_to_expand=['fruits'],
                            dummy_na=True,
                            fill_value=0.0)
    expander.fit(data_raw)
    arr = expander._expand_col(data_raw, 'fruits')
    arr_exp = np.array([[1., 0., 0.], [0., 0., 1.], [0., 1., 0.]])
    assert_almost_equal(arr, arr_exp)


def test_expand_col_numeric_no_dummy(data_raw):
    expander = DataFrameETL(cols_to_drop=['pid', 'djinn_type', 'animal'],
                            cols_to_expand=['fruits'],
                            dummy_na=False,
                            fill_value=np.nan)
    expander.fit(data_raw)
    arr = expander._expand_col(data_raw, 'fruits')
    arr_exp = np.array([[1., 0.], [np.nan, np.nan], [0., 1.]])
    assert_almost_equal(arr, arr_exp)


def test_expand_col_few_levels(data_few_levels, few_levels_expected):
    expander = DataFrameETL(cols_to_expand=['pid', 'fruits', 'animal'],
                            dummy_na=True,
                            fill_value=99.)
    expander.fit(data_few_levels)
    arr = expander._expand_col(data_few_levels, 'pid')
    expected_array = np.asarray(few_levels_expected[['pid_NaN', 'pid_a',
                                                     'pid_NaN_Sentinel']])
    assert_almost_equal(arr, expected_array)

    arr = expander._expand_col(data_few_levels, 'fruits')
    expected_array = np.asarray(few_levels_expected[['fruits_1.0',
                                                     'fruits_NaN']])
    assert_almost_equal(arr, expected_array)

    arr = expander._expand_col(data_few_levels, 'animal')
    expected_array = np.asarray(few_levels_expected[['animal_cat',
                                                     'animal_NaN']])
    assert_almost_equal(arr, expected_array)


def test_expand_col_few_levels_no_dummy(data_few_levels, few_levels_expected):
    expander = DataFrameETL(cols_to_expand=['pid', 'fruits', 'animal'],
                            dummy_na=False,
                            fill_value=99.)
    expander.fit(data_few_levels)
    arr = expander._expand_col(data_few_levels, 'pid')
    expected_array = np.asarray(few_levels_expected[['pid_NaN', 'pid_a']])
    assert_almost_equal(arr, expected_array)

    arr = expander._expand_col(data_few_levels, 'fruits')
    expected_array = np.asarray(few_levels_expected[['fruits_1.0']])
    assert_almost_equal(arr, expected_array)

    arr = expander._expand_col(data_few_levels, 'animal')
    expected_array = np.asarray(few_levels_expected[['animal_cat']])
    assert_almost_equal(arr, expected_array)


def test_fit_exc(dataframe_expected):
    expander = DataFrameETL()
    arr = np.asarray(dataframe_expected)
    with pytest.raises(TypeError):
        expander.fit(arr)


def test_fit_auto(data_raw, levels_dict):
    # test auto identification of columns to expand
    expander = DataFrameETL(cols_to_drop=['fruits'],
                            cols_to_expand='auto',
                            dummy_na=True)
    expander.fit(data_raw)
    cols_expected = ['pid_a', 'pid_b', 'pid_c', 'pid_NaN',
                     'djinn_type_effrit', 'djinn_type_marid',
                     'djinn_type_sila', 'djinn_type_NaN', 'age',
                     'animal_cat', 'animal_dog', 'animal_NaN']
    assert expander.levels_ == levels_dict
    assert expander.columns_ == cols_expected
    assert expander.required_columns_ == ['pid', 'djinn_type',
                                          'age', 'animal']


def test_fit_none(data_raw):
    # No dropping, no expansion
    expander = DataFrameETL(cols_to_drop=None,
                            cols_to_expand=None)
    expander.fit(data_raw)
    # column names should be identical to dataframe names
    data_cnames = data_raw.columns.tolist()
    assert expander.columns_ == data_cnames
    assert expander.required_columns_ == data_cnames


def test_fit_drop_only(data_raw):
    # test dropping columns but not expanding them
    expander = DataFrameETL(cols_to_drop=['fruits'],
                            cols_to_expand=None)
    expander.fit(data_raw)
    # both lists will be the same, since no cols were expanded
    assert expander.columns_ == ['pid', 'djinn_type', 'age', 'animal']
    assert expander.required_columns_ == ['pid', 'djinn_type',
                                          'age', 'animal']


def test_fit_list(data_raw, levels_dict):
    # test that fit handles list of cols to expand correctly
    expander = DataFrameETL(cols_to_drop=['fruits'],
                            cols_to_expand=['pid', 'djinn_type', 'animal'],
                            dummy_na=True)
    expander.fit(data_raw)
    cols_expected = ['pid_a', 'pid_b', 'pid_c', 'pid_NaN',
                     'djinn_type_effrit', 'djinn_type_marid',
                     'djinn_type_sila', 'djinn_type_NaN', 'age',
                     'animal_cat', 'animal_dog', 'animal_NaN']
    assert expander.levels_ == levels_dict
    assert expander.columns_ == cols_expected
    assert expander.required_columns_ == ['pid', 'djinn_type',
                                          'age', 'animal']


def test_fit_list_numeric(data_raw, levels_dict_numeric):
    expander = DataFrameETL(cols_to_drop=['djinn_type', 'animal'],
                            cols_to_expand=['pid', 'fruits'],
                            dummy_na=True)
    expander.fit(data_raw)
    assert expander.levels_ == levels_dict_numeric
    cols_numeric = ['pid_a', 'pid_b', 'pid_c', 'pid_NaN', 'fruits_1.0',
                    'fruits_3.0', 'fruits_NaN', 'age']
    assert expander.columns_ == cols_numeric
    assert expander.required_columns_ == ['pid', 'fruits', 'age']


def test_fit_with_nan_col(data_raw, levels_dict):
    # test that fit handles all-nan columns correctly
    data_raw['nantastic'] = pd.Series([np.NaN] * 3)
    expander = DataFrameETL(cols_to_drop=['fruits'],
                            cols_to_expand=['pid', 'djinn_type', 'animal'],
                            dummy_na=True,
                            check_null_cols='warn')
    with warnings.catch_warnings(record=True) as fit_w:
        expander.fit(data_raw)
        cols_expected = ['pid_a', 'pid_b', 'pid_c', 'pid_NaN',
                         'djinn_type_effrit', 'djinn_type_marid',
                         'djinn_type_sila', 'djinn_type_NaN', 'age',
                         'animal_cat', 'animal_dog', 'animal_NaN']
        assert expander.levels_ == levels_dict
        assert expander.columns_ == cols_expected
        assert expander.required_columns_ == ['pid', 'djinn_type',
                                              'age', 'animal']
        assert len(fit_w) == 1
        assert issubclass(fit_w[-1].category, UserWarning)
        assert expander._cols_to_drop == ['fruits', 'nantastic']

    with pytest.raises(RuntimeError):
        expander.check_null_cols = 'raise'
        expander.fit(data_raw)


def test_transform_notfitted(data_raw):
    expander = DataFrameETL()
    with pytest.raises(NotFittedError):
        expander.transform(data_raw)


def test_transform(data_raw, dataframe_expected):
    expander = DataFrameETL(cols_to_drop=['pid'],
                            cols_to_expand=['djinn_type', 'fruits', 'animal'],
                            dummy_na=True)
    expander.fit(data_raw)
    arr = expander.transform(data_raw)

    expected_array = np.asarray(dataframe_expected)
    assert arr.shape == expected_array.shape
    assert_almost_equal(arr, expected_array)


def test_transform_no_dummy(data_raw, dataframe_expected):
    expander = DataFrameETL(cols_to_drop=['pid'],
                            cols_to_expand=['djinn_type', 'fruits', 'animal'],
                            dummy_na=False)
    expander.fit(data_raw)
    arr = expander.transform(data_raw)

    # drop nan columns from expected data
    dataframe_expected.pop('djinn_type_NaN')
    dataframe_expected.pop('fruits_NaN')
    dataframe_expected.pop('animal_NaN')
    expected_array = np.asarray(dataframe_expected)
    assert arr.shape == expected_array.shape
    assert_almost_equal(arr, expected_array)


def test_transform_dataframe(data_raw, dataframe_expected):
    expander = DataFrameETL(cols_to_drop=['pid'],
                            cols_to_expand=['djinn_type', 'fruits', 'animal'],
                            dummy_na=True,
                            dataframe_output=True)
    expander.fit(data_raw)
    df = expander.transform(data_raw)
    assert df.shape == dataframe_expected.shape
    assert df.equals(dataframe_expected)


def test_transform_dataframe_no_dummy(data_raw, dataframe_expected):
    expander = DataFrameETL(cols_to_drop=['pid'],
                            cols_to_expand=['djinn_type', 'fruits', 'animal'],
                            dummy_na=False,
                            dataframe_output=True)
    expander.fit(data_raw)
    df = expander.transform(data_raw)

    # drop nan columns from expected data
    dataframe_expected.pop('djinn_type_NaN')
    dataframe_expected.pop('fruits_NaN')
    dataframe_expected.pop('animal_NaN')
    assert df.shape == dataframe_expected.shape
    assert df.equals(dataframe_expected)


def test_transform_two_levels(data_few_levels, few_levels_expected):
    expander = DataFrameETL(cols_to_expand=['pid', 'fruits', 'animal'],
                            dummy_na=True,
                            fill_value=99.,
                            dataframe_output=True)
    expander.fit(data_few_levels)
    df = expander.transform(data_few_levels)
    assert df.shape == few_levels_expected.shape
    assert df.equals(few_levels_expected)


def test_transform_two_levels_no_dummy(data_few_levels, few_levels_expected):
    expander = DataFrameETL(cols_to_expand=['pid', 'fruits', 'animal'],
                            dummy_na=False,
                            fill_value=99.,
                            dataframe_output=True)
    expander.fit(data_few_levels)
    df = expander.transform(data_few_levels)
    few_levels_expected.pop('pid_NaN_Sentinel')
    few_levels_expected.pop('fruits_NaN')
    few_levels_expected.pop('animal_NaN')
    assert df.shape == few_levels_expected.shape
    assert df.equals(few_levels_expected)


def test_transform_reuse_transformer(data_raw, data_raw_2,
                                     dataframe_2_expected):
    expander = DataFrameETL(cols_to_expand=['pid', 'fruits'],
                            cols_to_drop=['djinn_type', 'age', 'animal'],
                            dummy_na=True,
                            dataframe_output=True)
    expander.fit(data_raw)
    df = expander.transform(data_raw_2)
    assert df.equals(dataframe_2_expected)


def test_transform_preserve_col_order(data_raw, data_raw_2,
                                      dataframe_2_expected):
    expander = DataFrameETL(cols_to_expand=['pid', 'fruits'],
                            cols_to_drop=['djinn_type', 'age', 'animal'],
                            dummy_na=True,
                            dataframe_output=True)
    expander.fit(data_raw)
    # swap col order for second data file
    data_raw_2 = data_raw_2[['fruits', 'age', 'djinn_type', 'pid', 'animal']]
    df = expander.transform(data_raw_2)
    assert df.equals(dataframe_2_expected)


def test_transform_bad_expand_col(data_raw, dataframe_expected):
    expander = DataFrameETL(cols_to_drop=['pid'],
                            cols_to_expand=['djinn_type', 'fruits',
                                            'animal', 'not_in_df'],
                            dummy_na=True)
    expander.fit(data_raw)
    arr = expander.transform(data_raw)

    expected_array = np.asarray(dataframe_expected)
    assert arr.shape == expected_array.shape
    assert_almost_equal(arr, expected_array)


def test_refit(data_raw, data_raw_2):
    expander = DataFrameETL(cols_to_drop=['pid', 'age', 'djinn_type'],
                            cols_to_expand=['fruits', 'animal'],
                            dataframe_output=True,
                            dummy_na=True)
    expander.fit(data_raw)
    df = expander.transform(data_raw)
    df_expected = pd.concat([
        pd.Series([1., 0., 0.], dtype='float32', name='fruits_1.0'),
        pd.Series([0., 0., 1.], dtype='float32', name='fruits_3.0'),
        pd.Series([0., 1., 0.], dtype='float32', name='fruits_NaN'),
        pd.Series([1., 0., 0.], dtype='float32', name='animal_cat'),
        pd.Series([0., 1., 0.], dtype='float32', name='animal_dog'),
        pd.Series([0., 0., 1.], dtype='float32', name='animal_NaN'),
    ], axis=1)
    assert df.equals(df_expected)

    expander.fit(data_raw_2)
    df2 = expander.transform(data_raw_2)
    df_expected_2 = pd.concat([
        pd.Series([1., 0., 0.], dtype='float32', name='fruits_-99999.0'),
        pd.Series([0., 0., 1.], dtype='float32', name='fruits_1.0'),
        pd.Series([0., 1., 0.], dtype='float32', name='fruits_NaN'),
        pd.Series([1., 0., 0.], dtype='float32', name='animal_cat'),
        pd.Series([0., 1., 0.], dtype='float32', name='animal_dog'),
        pd.Series([0., 0., 1.], dtype='float32', name='animal_NaN'),
    ], axis=1)
    assert df2.equals(df_expected_2)


def test_dataframe_output_with_index():
    # DataFrame output should preserve the index of the input
    df = pd.DataFrame({'a': [1, 2, 3], 'b': ['x', 'y', 'x']}, index=[11, 8, 0])
    df2 = DataFrameETL(dataframe_output=True, dummy_na=False).fit_transform(df)

    df_exp = pd.DataFrame({'a': [1, 2, 3], 'b_x': [1, 0, 1], 'b_y': [0, 1, 0]},
                          index=[11, 8, 0], dtype='float32')

    assert df2.equals(df_exp)


def test_pickle(data_raw):
    expander = DataFrameETL(cols_to_drop=['pid'],
                            cols_to_expand=['djinn_type', 'fruits', 'animal'],
                            dummy_na=True)
    expected_array = expander.fit_transform(data_raw)
    # pickle the transformer
    buff = io.BytesIO()
    pickle.dump(expander, buff)
    buff.seek(0)
    # transform data after unpickling transformer
    expander = pickle.load(buff)

    arr = expander.transform(data_raw)
    assert arr.shape == expected_array.shape
    assert_almost_equal(arr, expected_array)


def test_categorical_looks_like_int():
    # Verify that the right thing happens if the fit DataFrame has a
    # categorical with integer categories
    raw = pd.concat([
            pd.Series([1.0, np.NaN, 3.0], dtype='float', name='fruits'),
            pd.Series([500, 1000, 1000], dtype='category', name='intcat'),
        ], axis=1)
    expander = DataFrameETL(cols_to_expand='auto', dummy_na=True)
    tfm = expander.fit_transform(raw)
    exp = np.array([[1, 1, 0, 0],
                    [np.nan, 0, 1, 0],
                    [3, 0, 1, 0]])
    assert_almost_equal(tfm, exp)
    assert expander.columns_ == \
        ['fruits', 'intcat_500', 'intcat_1000', 'intcat_NaN']


def test_categorical_mixed_type_levels():
    # Use a category with both strings and integers in its categories
    raw = pd.concat([
            pd.Series([1.0, np.NaN, 3.0], dtype='float', name='fruits'),
            pd.Series([500, np.NaN, 'cat'], dtype='category', name='mixed'),
        ], axis=1)
    expander = DataFrameETL(cols_to_expand='auto', dummy_na=False)
    tfm = expander.fit_transform(raw)
    exp = np.array([[1, 1, 0],
                    [np.nan, 0, 0],
                    [3, 0, 1]])
    assert_almost_equal(tfm, exp)
    assert expander.columns_ == ['fruits', 'mixed_500', 'mixed_cat']



from __future__ import print_function
from __future__ import division

import logging
import uuid
import warnings
from itertools import chain

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.exceptions import NotFittedError

log = logging.getLogger(__name__)


class DataFrameETL(BaseEstimator, TransformerMixin):
    """Performs ETL on a dataframe, using the following steps:
    - dropping specified columns
    - performing categorical expansion on a specified set of columns
    - converting the dataframe to a numpy array or dataframe of float32s.

    Parameters
    ----------
    cols_to_drop : list[str], optional (default: None)
        List of columns to drop from the dataframe.
    cols_to_expand: {"auto", None, list[str]}, (default: 'auto')
        Which columns should be expanded:
        - "auto": All non-numeric columns will be expanded.
        - None: no columns will be expanded.
        - list[str]: list of column names to expand.
    dummy_na : bool, (default: True)
        Add a column to indicate missing values.
    fill_value : {float, np.nan}, (default: 0.0)
        The value to fill for missing values with in expanded columns.
        Can be a float or np.nan.
    dataframe_output : bool (default: False)
        If True, ETL output is a pd.DataFrame instead of a np.Array.
    check_null_cols : {None, False, 'raise', 'warn'} (default: False)
        How columns of all nulls should be handled:
        - None or False: do not check for null columns (best performance
          during `fit`).
        - 'raise': raises a RuntimeError if null columns are found
        - 'warn': issues a warning if null columns are found

    Attributes
    ----------
    levels_, dict
        Dictionary created which stores the transformation. Each key is a
        column and each value is an ordered list of values corresponding
        to the categorical representation of the column in question.
    required_columns_, list[str]
        List of column names without cols_to_drop, but before expanding
        cols_to_expand. This is the minimal list of columns required to
        be in the input data when .transform is called.
    columns_, list[str]
        List of final column names in order
    """
    expansion_warn_threshold = 500  # Warn when expanding this many categories

    def __init__(self,
                 cols_to_drop=None,
                 cols_to_expand='auto',
                 dummy_na=True,
                 fill_value=0.0,
                 dataframe_output=False,
                 check_null_cols=False):
        self.cols_to_drop = cols_to_drop
        self.cols_to_expand = cols_to_expand
        self.dummy_na = dummy_na
        self.fill_value = fill_value
        self.dataframe_output = dataframe_output
        self.check_null_cols = check_null_cols

    def _flag_nulls(self, X, cols_to_drop):
        null_cols = [col for col in X if
                     col not in cols_to_drop and
                     pd.isnull(X[col].values[0]) and
                     X[col].first_valid_index() is None]
        if len(null_cols) > 0:
            if self.check_null_cols == 'warn':
                warnings.warn('The following columns contain only nulls '
                              'and will be dropped: ' + str(null_cols),
                              UserWarning)
            elif self.check_null_cols == 'raise':
                raise RuntimeError('The following columns contain only '
                                   'nulls: ' + str(null_cols))
            else:
                raise ValueError('DataFrameETL.check_null_cols must be '
                                 'one of the following: [None, False, '
                                 '"raise", or "warn"]')
        return cols_to_drop + null_cols

    def _check_sentinels(self, X):
        """Replace default sentinel with random values if the default appears
        in the data."""
        vals = chain.from_iterable(pd.unique(X[col]) for
                                   col in self._cols_to_expand)

        while any(u == self._nan_sentinel for u in vals):
            self._nan_sentinel = uuid.uuid4().hex

    def _create_levels(self, X):
        """Create levels for each column in cols_to_expand."""
        levels = {}
        warn_list = {}
        # get a list of categories when the column is cast to
        # dtype category
        # levels are sorted by default
        for col in self._cols_to_expand:
            levels[col] = X[col].astype('category').cat.categories.tolist()
            if (self.expansion_warn_threshold and
                    len(levels[col]) >= self.expansion_warn_threshold):
                warn_list[col] = len(levels[col])
            # if there are nans, we will be replacing them with a sentinel,
            # so add the sentinel as a level explicitly
            # Note that even if we don't include a dummy_na column, we still
            # need to keep track of missing values internally for fill_value
            if self.dummy_na or any(X[col].isnull()):
                levels[col].extend([self._nan_sentinel])
        log.debug("Categories (including nulls) for each column: %s",
                  "; ".join('"%s": %d' % (c, len(l))
                            for c, l in levels.items()))
        if warn_list:
            warnings.warn("The following categorical column(s) have a large "
                          "number of categories. Are you sure you wish to "
                          "convert them to binary indicators?\n%s" %
                          ("; ".join(['"%s": %d categories' % (c, l)
                                      for c, l in warn_list.items()])),
                          RuntimeWarning)
        return levels

    def _create_col_names(self, X):
        """Identify levels and build an ordered list of final column names."""
        # get unexpanded columns, remove any that need dropping
        unexpanded_cnames = [col for col in X.columns.tolist() if
                             col not in self._cols_to_drop]

        # replace unexpanded colnames with a list of expanded colnames,
        # to maintain column order
        cnames = []
        for col in unexpanded_cnames:
            if col in self._cols_to_expand:
                col_levels = self.levels_[col]
                if self.dummy_na:
                    # avoid exposing the sentinel to the user by replacing
                    # it with 'NaN'. If 'NaN' is already a level, use the
                    # sentinel to prevent column name duplicates.
                    if 'NaN' in col_levels:
                        expanded_names = ['%s_%s' % (col, self._nan_sentinel)
                                          if cat == self._nan_sentinel
                                          else '%s_%s' % (col, cat) for cat in
                                          col_levels]
                    else:
                        expanded_names = ['%s_NaN' % (col)
                                          if cat == self._nan_sentinel
                                          else '%s_%s' % (col, cat) for cat in
                                          col_levels]
                else:
                    # if the final data frame will not have a dummy na column,
                    # don't include it in the unexpanded list.
                    expanded_names = ['%s_%s' % (col, cat) for cat in
                                      col_levels if cat != self._nan_sentinel]
                cnames.extend(expanded_names)
            else:
                cnames.append(col)

        return cnames, unexpanded_cnames

    def _expand_col(self, X, col):
        """Perform categorical expansion on a single column."""
        # Convert the input to a categorical and fill missing
        # values with our sentinel. If the input has sentinels already,
        # then it's a category not seen during the fit and should
        # be ignored in the expansion.
        catcol = X[col].astype('category')
        if self._nan_sentinel not in catcol.cat.categories:
            catcol = catcol.cat.add_categories(self._nan_sentinel)
            sentinel_entries = None
        else:
            sentinel_entries = (catcol == self._nan_sentinel)
        catcol = catcol.fillna(self._nan_sentinel)

        # One-hot-encode the array, using the levels seen during fit.
        # When we expand, the only missing values will be categories
        # not seen during fit. We ignore those when encoding
        # by using `dummy_na=False`.
        newcat = catcol.cat.set_categories(self.levels_[col])
        expanded_array = pd.get_dummies(newcat, dummy_na=False)
        expanded_array = expanded_array.values.astype('float32')

        is_nan = (newcat == self._nan_sentinel)
        if is_nan.any():
            expanded_array[is_nan, :-1] = self.fill_value
            if not self.dummy_na:
                # Drop the last column, which is the nan column
                expanded_array = expanded_array[:, :-1]

        # replace sentinel entries which were treated like nans to
        # 0 in the final array
        if sentinel_entries is not None:
            expanded_array[sentinel_entries, :] = 0.0
        return expanded_array

    def fit(self, X, y=None):
        """Fit the ETL pipeline.

        Parameters
        ----------
        X : pd.DataFrame
            Training or test data.
        y : numpy array of shape [n_samples]
            Ignored by this class.

        Returns
        -------
        self : returns an instance of DataFrameETL
        """
        if not isinstance(X, pd.DataFrame):
            raise TypeError("ETL transformer must be fit to a dataframe.")

        # set default values
        #  _nan_sentinel is a temporary sentinel for np.nan values
        self._nan_sentinel = 'NaN_Sentinel'
        self.levels_ = {}
        if self.cols_to_drop is None:
            self._cols_to_drop = []
        else:
            self._cols_to_drop = self.cols_to_drop
        # Remove any columns which are all np.nan
        if self.check_null_cols:
            self._cols_to_drop = self._flag_nulls(X, self._cols_to_drop)

        # If None, skip fit step, since we won't do any expansion
        if self.cols_to_expand is None:
            self._cols_to_expand = []
        else:
            if self.cols_to_expand == 'auto':
                self._cols_to_expand = X.columns[
                    (X.dtypes == 'object') | (X.dtypes == 'category')].tolist()
            else:
                self._cols_to_expand = [c for c in self.cols_to_expand if
                                        c in X.columns]
            self._cols_to_expand = [c for c in self._cols_to_expand if
                                    c not in self._cols_to_drop]
            log.debug("There are %d column(s) to expand.",
                      len(self._cols_to_expand))
            # Update sentinels if the defaults are in the dataframe
            self._check_sentinels(X)
            self.levels_ = self._create_levels(X)

        # Get colummn names in order
        self.columns_, self.required_columns_ = self._create_col_names(X)

        return self

    def transform(self, X):
        """Create an expanded array or DataFrame and fill it with
        transformed data in chunks, for memory efficiency.

        Parameters
        ----------
        X : pd.DataFrame
            Training or test data.

        Returns
        -------
        np.Array or pd.DataFrame
            Original dataframe with dropped columns and expansion of
            categorical variables, converted to an array or dataframe
            of float32s.
        """

        # If columns_ attribute isn't present, it hasn't been fitted
        if not hasattr(self, 'columns_'):
            raise NotFittedError('This DataFrameETL instance is '
                                 'not fitted yet',)

        log.debug("The input data for transformation have shape %s. "
                  "They will be expanded to shape (%d, %d).",
                  str(X.shape), X.shape[0], len(self.columns_))
        if self.dataframe_output:
            # preallocate a dataframe
            X_new = pd.DataFrame(index=X.index, columns=self.columns_)
            # column index
            i = 0
            for col in self.required_columns_:
                if col in self._cols_to_expand:
                    # assigning this to a temp variable so we can
                    # figure out its shape
                    expanded = self._expand_col(X, col)
                    # this loop is to prevent type coercions from
                    # assigning values to multiple pandas columns
                    # at once
                    for j in range(expanded.shape[1]):
                        X_new.iloc[:, i + j] = expanded[:, j]
                    i += expanded.shape[1]
                else:
                    # put the column in the array
                    X_new.iloc[:, i] = X[col].astype('float32')
                    i += 1
        else:
            # preallocate an array
            ncol = len(self.columns_)
            X_new = np.empty([X.shape[0], ncol], dtype='float32')
            # column index
            i = 0
            for col in self.required_columns_:
                if col in self._cols_to_expand:
                    # assigning this to a temp variable so we can
                    # figure out its shape
                    expanded = self._expand_col(X, col)
                    X_new[:, i:(i + expanded.shape[1])] = expanded
                    i += expanded.shape[1]
                else:
                    # put the column in the array
                    X_new[:, i] = X[col].astype('float32')
                    i += 1

        return X_new



from __future__ import print_function
from __future__ import division

import pytest
import numpy as np
from sklearn.exceptions import ConvergenceWarning

from civismlext.nonnegative import NonNegativeLinearRegression
from civismlext.nonnegative import _rescale_data

rng = np.random.RandomState(17)


def test_rescale_data():
    """Copied from sklearn/linear_model/tests/test_base.py
    """
    n_samples = 200
    n_features = 2

    sample_weight = 1.0 + rng.rand(n_samples)
    X = rng.rand(n_samples, n_features)
    y = rng.rand(n_samples)
    rescaled_X, rescaled_y = _rescale_data(X, y, sample_weight)
    rescaled_X2 = X * np.sqrt(sample_weight)[:, np.newaxis]
    rescaled_y2 = y * np.sqrt(sample_weight)
    np.testing.assert_allclose(rescaled_X, rescaled_X2)
    np.testing.assert_allclose(rescaled_y, rescaled_y2)


def test_nonneg_linreg():
    X = np.array([[0, 1], [1, 0], [0, 0], [1, 1]])

    nnlr = NonNegativeLinearRegression(fit_intercept=True)
    nnlr.fit(X, X.dot([1, 2]) + 2)

    np.testing.assert_allclose(nnlr.coef_, [1, 2])
    np.testing.assert_allclose(nnlr.intercept_, [2])
    np.testing.assert_allclose(nnlr.predict(np.array([[1, 1]])), [5])


def test_zero_coef_exception():
    with pytest.raises(ConvergenceWarning):
        X = np.array([[0, 1], [1, 0], [0, 0], [1, 1]])

        nnlr = NonNegativeLinearRegression(fit_intercept=True)
        nnlr.fit(X, X.dot([-1, -1]) + -1)


def test_preprocess_data():
    """Copied from sklearn/linear_model/tests/test_base.py, with small
    modifications.
    """
    n_samples = 200
    n_features = 2
    X = rng.rand(n_samples, n_features)
    y = rng.rand(n_samples)
    expected_X_mean = np.mean(X, axis=0)
    expected_X_norm = np.std(X, axis=0) * np.sqrt(X.shape[0])
    expected_y_mean = np.mean(y, axis=0)
    nnlr = NonNegativeLinearRegression()

    Xt, yt, X_mean, y_mean, X_norm = \
        nnlr._preprocess_data(X, y, fit_intercept=False, normalize=False)
    np.testing.assert_allclose(X_mean, np.zeros(n_features))
    np.testing.assert_allclose(y_mean, 0)
    np.testing.assert_allclose(X_norm, np.ones(n_features))
    np.testing.assert_allclose(Xt, X)
    np.testing.assert_allclose(yt, y)

    Xt, yt, X_mean, y_mean, X_norm = \
        nnlr._preprocess_data(X, y, fit_intercept=True, normalize=False)
    np.testing.assert_allclose(X_mean, expected_X_mean)
    np.testing.assert_allclose(y_mean, expected_y_mean)
    np.testing.assert_allclose(X_norm, np.ones(n_features))
    np.testing.assert_allclose(Xt, X - expected_X_mean)
    np.testing.assert_allclose(yt, y - expected_y_mean)

    Xt, yt, X_mean, y_mean, X_norm = \
        nnlr._preprocess_data(X, y, fit_intercept=True, normalize=True)
    np.testing.assert_allclose(X_mean, expected_X_mean)
    np.testing.assert_allclose(y_mean, expected_y_mean)
    np.testing.assert_allclose(X_norm, expected_X_norm)
    np.testing.assert_allclose(Xt, (X - expected_X_mean) / expected_X_norm)
    np.testing.assert_allclose(yt, y - expected_y_mean)


def test_preprocess_data_weighted():
    """Copied from sklearn/linear_model/tests/test_base.py, with small
    modifications.
    """
    n_samples = 200
    n_features = 2
    X = rng.rand(n_samples, n_features)
    y = rng.rand(n_samples)
    sample_weight = rng.rand(n_samples)
    expected_X_mean = np.average(X, axis=0, weights=sample_weight)
    expected_y_mean = np.average(y, axis=0, weights=sample_weight)
    nnlr = NonNegativeLinearRegression()

    # XXX: if normalize=True, should we expect a weighted standard deviation?
    #      Currently not weighted, but calculated with respect to weighted mean
    expected_X_norm = (np.sqrt(X.shape[0]) *
                       np.mean((X - expected_X_mean) ** 2, axis=0) ** .5)

    Xt, yt, X_mean, y_mean, X_norm = \
        nnlr._preprocess_data(X, y, fit_intercept=True, normalize=False,
                              sample_weight=sample_weight)
    np.testing.assert_allclose(X_mean, expected_X_mean)
    np.testing.assert_allclose(y_mean, expected_y_mean)
    np.testing.assert_allclose(X_norm, np.ones(n_features))
    np.testing.assert_allclose(Xt, X - expected_X_mean)
    np.testing.assert_allclose(yt, y - expected_y_mean)

    Xt, yt, X_mean, y_mean, X_norm = \
        nnlr._preprocess_data(X, y, fit_intercept=True, normalize=True,
                              sample_weight=sample_weight)
    np.testing.assert_allclose(X_mean, expected_X_mean)
    np.testing.assert_allclose(y_mean, expected_y_mean)
    np.testing.assert_allclose(X_norm, expected_X_norm)
    np.testing.assert_allclose(Xt, (X - expected_X_mean) / expected_X_norm)
    np.testing.assert_allclose(yt, y - expected_y_mean)



from __future__ import print_function
from __future__ import division

from collections import defaultdict
from functools import partial
import copy
import itertools
import logging

import numpy as np
from scipy.stats import rankdata

from sklearn.model_selection import ParameterSampler
from sklearn.model_selection._search import BaseSearchCV
from sklearn.base import is_classifier, clone
from sklearn.model_selection import check_cv

# I don't want this, but fine.
from sklearn.model_selection._validation import _fit_and_score

from sklearn.externals.joblib import Parallel, delayed
from sklearn.utils.fixes import MaskedArray
from sklearn.utils.validation import indexable
from sklearn.metrics.scorer import check_scoring
from sklearn.utils import check_random_state

log = logging.getLogger(__name__)


def hyperband_num_per_run(eta, R, Rmin):
    num = 0
    smax = int(np.floor(np.log(R / Rmin) / np.log(eta)))
    B = (smax + 1.0) * R
    for s in range(smax, -1, -1):
        n = int(np.ceil(B / R * np.power(eta, s) / (s + 1.0)))
        T = [0] * n
        for i in range(0, s + 1):
            n_i = int(np.floor(n / np.power(eta, i)))
            num_to_keep = int(np.floor(n_i / eta))
            num += len(T)
            T = T[0:num_to_keep]

    return num


class HyperbandSearchCV(BaseSearchCV):
    """Hyperband search on hyper parameters.

    If all parameters are presented as a list,
    sampling without replacement is performed. If at least one parameter
    is given as a distribution, sampling with replacement is used.

    It is highly recommended to use continuous distributions for continuous
    parameters.

    Parameters
    ----------
    estimator : estimator object.
        A object of that type is instantiated for each grid point.
        This is assumed to implement the scikit-learn estimator interface.
        Either estimator needs to provide a ``score`` function,
        or ``scoring`` must be passed.
    param_distributions : dict
        Dictionary with parameters names (string) as keys and distributions
        or lists of parameters to try. Distributions must provide a ``rvs``
        method for sampling (such as those from scipy.stats.distributions).
        If a list is given, it is sampled uniformly.
    cost_parameter_max : dict
        Dictionary with the cost parameter name as a key and the maximum
        cost as the value. The cost parameter is the maximum cost per randomly
        selected hyperparameter configuration. Typicaly this parameter would
        be the number of iterations for SGD-like methods or the number of
        trees in an ensemble.
    cost_parameter_min : dict, optional
        Dictionary with the cost parameter name as a key and the minimum
        cost as the value. Defaults to 1 if not given. This parameter is the
        minimum of the cost parameter. This option can be used to force
        hyperband to use, say, at least 100 trees in an ensemble method in
        order to limit the hyperband to reasonable choices in hyperparameter
        space.
    eta : int, default=3
        One over the fraction of configurations that are discarded in each
        round of parameter searching.
    scoring : string, callable or None, default=None
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y)``.
        If ``None``, the ``score`` method of the estimator is used.
    n_jobs : int, default=1
        Number of jobs to run in parallel.
    pre_dispatch : int, or string, optional
        Controls the number of jobs that get dispatched during parallel
        execution. Reducing this number can be useful to avoid an
        explosion of memory consumption when more jobs get dispatched
        than CPUs can process. This parameter can be:
            - None, in which case all the jobs are immediately
              created and spawned. Use this for lightweight and
              fast-running jobs, to avoid delays due to on-demand
              spawning of the jobs
            - An int, giving the exact number of total jobs that are
              spawned
            - A string, giving an expression as a function of n_jobs,
              as in '2*n_jobs'
    iid : boolean, default=True
        If True, the data is assumed to be identically distributed across
        the folds, and the loss minimized is the total loss per sample,
        and not the mean loss across the folds.
    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:
          - None, to use the default 3-fold cross validation,
          - integer, to specify the number of folds in a `(Stratified)KFold`,
          - An object to be used as a cross-validation generator.
          - An iterable yielding train, test splits.
        For integer/None inputs, if the estimator is a classifier and ``y`` is
        either binary or multiclass, :class:`StratifiedKFold` is used. In all
        other cases, :class:`KFold` is used.
        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.
    refit : boolean, default=True
        Refit the best estimator with the entire dataset.
        If "False", it is impossible to make predictions using
        this HyperbandSearchCV instance after fitting.
    verbose : integer
        Controls the verbosity: the higher, the more messages.
    random_state : int or RandomState
        Pseudo random number generator state used for random uniform sampling
        from lists of possible values instead of scipy.stats distributions.
    error_score : 'raise' (default) or numeric
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised. If a numeric value is given,
        FitFailedWarning is raised. This parameter does not affect the refit
        step, which will always raise the error.
    return_train_score : boolean, default=True
        If ``'False'``, the ``cv_results_`` attribute will not include training
        scores.

    Attributes
    ----------
    cv_results_ : dict of numpy (masked) ndarrays
        A dict with keys as column headers and values as columns, that can be
        imported into a pandas ``DataFrame``.
        For instance the below given table
        +--------------+-------------+-------------------+---+---------------+
        | param_kernel | param_gamma | split0_test_score |...|rank_test_score|
        +==============+=============+===================+===+===============+
        |    'rbf'     |     0.1     |        0.8        |...|       2       |
        +--------------+-------------+-------------------+---+---------------+
        |    'rbf'     |     0.2     |        0.9        |...|       1       |
        +--------------+-------------+-------------------+---+---------------+
        |    'rbf'     |     0.3     |        0.7        |...|       1       |
        +--------------+-------------+-------------------+---+---------------+
        will be represented by a ``cv_results_`` dict of::
            {
            'param_kernel' : masked_array(data = ['rbf', 'rbf', 'rbf'],
                                          mask = False),
            'param_gamma'  : masked_array(data = [0.1 0.2 0.3], mask = False),
            'split0_test_score'  : [0.8, 0.9, 0.7],
            'split1_test_score'  : [0.82, 0.5, 0.7],
            'mean_test_score'    : [0.81, 0.7, 0.7],
            'std_test_score'     : [0.02, 0.2, 0.],
            'rank_test_score'    : [3, 1, 1],
            'split0_train_score' : [0.8, 0.9, 0.7],
            'split1_train_score' : [0.82, 0.5, 0.7],
            'mean_train_score'   : [0.81, 0.7, 0.7],
            'std_train_score'    : [0.03, 0.03, 0.04],
            'mean_fit_time'      : [0.73, 0.63, 0.43, 0.49],
            'std_fit_time'       : [0.01, 0.02, 0.01, 0.01],
            'mean_score_time'    : [0.007, 0.06, 0.04, 0.04],
            'std_score_time'     : [0.001, 0.002, 0.003, 0.005],
            'params' : [{'kernel' : 'rbf', 'gamma' : 0.1}, ...],
            }
        NOTE that the key ``'params'`` is used to store a list of parameter
        settings dict for all the parameter candidates.
        The ``mean_fit_time``, ``std_fit_time``, ``mean_score_time`` and
        ``std_score_time`` are all in seconds.
    best_estimator_ : estimator
        Estimator that was chosen by the search, i.e. estimator
        which gave highest score (or smallest loss if specified)
        on the left out data. Not available if refit=False.
    best_score_ : float
        Score of best_estimator on the left out data.
    best_params_ : dict
        Parameter setting that gave the best results on the hold out data.
    best_index_ : int
        The index (of the ``cv_results_`` arrays) which corresponds to the best
        candidate parameter setting.
        The dict at ``search.cv_results_['params'][search.best_index_]`` gives
        the parameter setting for the best model, that gives the highest
        mean score (``search.best_score_``).
    scorer_ : function
        Scorer function used on the held out data to choose the best
        parameters for the model.
    n_splits_ : int
        The number of cross-validation splits (folds/iterations).

    Notes
    -----
    The parameters selected are those that maximize the score of the held-out
    data, according to the scoring parameter.

    If `n_jobs` was set to a value higher than one, the data is copied for each
    parameter setting(and not `n_jobs` times). This is done for efficiency
    reasons if individual jobs take very little time, but may raise errors if
    the dataset is large and not enough memory is available.  A workaround in
    this case is to set `pre_dispatch`. Then, the memory is copied only
    `pre_dispatch` many times. A reasonable value for `pre_dispatch` is
    `2 * n_jobs`.

    See the original paper here https://arxiv.org/abs/1603.06560. The variable
    names in the code reflect the chosen names in the paper for clarity.

    Examples
    --------
    >>> rng = np.random.RandomState(seed=seed)
    >>> from scipy.stats import uniform
    >>> cv = KFold(
    >>>     n_splits=4,
    >>>     shuffle=True,
    >>>     random_state=rng.randint(1, 100000))
    >>> rf = RandomForestClassifier()
    >>> cvest = HyperbandSearchCV(
        rf,
        param_distributions={'criterion': ['gini', 'entropy'],
                             'max_depth': [None, 1, 4, 8, 16, 32],
                             'clf__max_features': uniform()},
        cost_parameter_max={'n_estimators': 1000},
        cost_parameter_min={'n_estimators': 100},
        eta=3,
        cv=cv,
        random_state=rng.randint(1, 100000),
        scoring='roc_auc')
    >>> cvest.fit(X, y)

    See Also
    --------
    :class:`RandomizedSearchCV`:
        Searches randomly over a predefined parameter space.
    :class:`ParameterSampler`:
        A generator over parameter settins, constructed from
        param_distributions.
    """

    def __init__(self, estimator, param_distributions, cost_parameter_max,
                 cost_parameter_min=None, eta=3,
                 scoring=None, n_jobs=1, iid=True, refit=True,
                 cv=None, verbose=0, pre_dispatch='2*n_jobs',
                 random_state=None, error_score='raise',
                 return_train_score=True):
        self.param_distributions = param_distributions
        self.cost_parameter_max = cost_parameter_max
        self.cost_parameter_min = cost_parameter_min
        self.eta = eta
        self.random_state = random_state
        super(HyperbandSearchCV, self).__init__(
            estimator=estimator, scoring=scoring,
            n_jobs=n_jobs, iid=iid, refit=refit, cv=cv, verbose=verbose,
            pre_dispatch=pre_dispatch, error_score=error_score,
            return_train_score=return_train_score)

    def fit(self, X, y=None, groups=None, **fit_params):
        """Run fit on the estimator with randomly drawn parameters.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Training vector, where n_samples in the number of samples and
            n_features is the number of features.
        y : array-like, shape = [n_samples] or [n_samples, n_output], optional
            Target relative to X for classification or regression;
            None for unsupervised learning.
        groups : array-like, with shape (n_samples,), optional
            Group labels for the samples used while splitting the dataset into
            train/test set.
        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of the estimator
        """
        estimator = self.estimator
        cv = check_cv(self.cv, y, classifier=is_classifier(estimator))
        self.scorer_ = check_scoring(self.estimator, scoring=self.scoring)
        self._random_state = check_random_state(self.random_state)
        X, y, groups = indexable(X, y, groups)
        n_splits = cv.get_n_splits(X, y, groups)
        R = list(self.cost_parameter_max.values())[0]

        if self.cost_parameter_min is None:
            Rmin = 1
        else:
            Rmin = list(self.cost_parameter_min.values())[0]

        n_candidates = hyperband_num_per_run(self.eta, R, Rmin)
        log.debug("Fitting %d folds for each of %d candidates, totalling "
                  "%d fits.", n_splits, n_candidates, n_candidates * n_splits)
        if self.verbose > 0:
            print("Fitting {0} folds for each of {1} candidates, totalling"
                  " {2} fits".format(n_splits, n_candidates,
                                     n_candidates * n_splits))

        base_estimator = clone(self.estimator)
        pre_dispatch = self.pre_dispatch

        cv_iter = list(cv.split(X, y, groups))

        out = []
        smax = int(np.floor(np.log(R / Rmin) / np.log(self.eta)))
        B = (smax + 1.0) * R

        # This code is hyperband, but I have swapped the order of the
        # inner and outer loops to expose more parallelism. Fun.
        Ts = []
        ns = []
        rs = []
        for s in range(smax, -1, -1):
            ns.append(int(np.ceil(B / R * np.power(self.eta, s) / (s + 1.0))))
            rs.append(int(R / np.power(self.eta, s)))
            Ts.append(list(ParameterSampler(
                self.param_distributions,
                ns[-1],
                random_state=self._random_state)))
        nums = copy.copy(ns)
        # these are the offsets to the hyperparameter configurations for
        # each value of s in the loop above
        # they get updated as the loop over the different rounds get run
        # below
        offsets = [0] + list(
            np.cumsum(np.array(nums) * n_splits).astype(int))[:-1]

        # iterate the maximum number of times for each resource budget
        # configuration.
        # If we should skip an interation, T will be an empty list
        for rnd in range(0, smax + 1):
            # set the costs for this round
            r_rnd = []
            for ind, s in enumerate(range(smax, -1, -1)):
                _r = int(rs[ind] * np.power(self.eta, rnd))
                r_rnd += [_r] * nums[ind]

            # run the jobs
            _jobs = []
            for parameters, _r in zip(
                    itertools.chain.from_iterable(Ts), r_rnd):
                _parameters = copy.deepcopy(parameters)
                _parameters.update(
                    {list(self.cost_parameter_max.keys())[0]: _r})
                for train, test in cv_iter:
                    _jobs.append(delayed(_fit_and_score)(
                        clone(base_estimator), X, y, self.scorer_,
                        train, test, self.verbose, _parameters,
                        fit_params=fit_params,
                        return_train_score=self.return_train_score,
                        return_n_test_samples=True,
                        return_times=True, return_parameters=True,
                        error_score=self.error_score))
            _out = Parallel(
                n_jobs=self.n_jobs, verbose=self.verbose,
                pre_dispatch=pre_dispatch)(_jobs)
            out += _out

            # now post-process
            new_Ts = []
            new_nums = []
            for ind, s in enumerate(range(smax, -1, -1)):
                n_i = int(np.floor(ns[ind] / np.power(self.eta, rnd)))
                num_to_keep = int(np.floor(n_i / self.eta))
                # keep for next round only if num_to_keep > 0 AND
                # the round after this round will be executed
                # in otherwords, you only need to cut the configurations
                # down by eta if you are going to test them in the next
                # round
                if num_to_keep > 0 and rnd < s:
                    _out_s = _out[
                        offsets[ind]:(offsets[ind] + nums[ind] * n_splits)]
                    results, _ = self._process_outputs(_out_s, n_splits)
                    sind = np.argsort(results["rank_test_score"])
                    msk = np.zeros(len(results['rank_test_score']))
                    msk[sind[0:num_to_keep]] = 1
                    msk = msk.astype(bool)
                    new_Ts.append(
                        [p for k, p in enumerate(results['params']) if msk[k]])
                    new_nums.append(num_to_keep)
                else:
                    new_Ts.append([])
                    new_nums.append(0)

            Ts = new_Ts
            nums = new_nums
            offsets = [0] + list(
                np.cumsum(np.array(nums) * n_splits).astype(int))[:-1]

        results, best_index = self._process_outputs(out, n_splits)
        self.cv_results_ = results
        self.best_index_ = best_index
        self.n_splits_ = n_splits
        self.multimetric_ = False
        if not hasattr(self, 'best_score_'):
            self.best_score_ = results['mean_test_score'][best_index]
        if not hasattr(self, 'best_params_'):
            self.best_params_ = results['params'][best_index]

        if self.refit:
            best_estimator = clone(self.estimator).set_params(
                **self.cv_results_['params'][self.best_index_])

            if y is not None:
                best_estimator.fit(X, y, **fit_params)
            else:
                best_estimator.fit(X, **fit_params)

            self.best_estimator_ = best_estimator

        return self

    def _store_results(
            self,
            results,
            n_splits,
            n_candidates,
            key_name,
            array,
            weights=None,
            splits=False,
            rank=False):
        """A small helper to store the scores/times to the cv_results_"""
        array = np.array(array, dtype=np.float64).reshape(n_candidates,
                                                          n_splits)
        if splits:
            for split_i in range(n_splits):
                results["split%d_%s"
                        % (split_i, key_name)] = array[:, split_i]

        array_means = np.average(array, axis=1, weights=weights)
        results['mean_%s' % key_name] = array_means
        # Weighted std is not directly available in numpy
        array_stds = np.sqrt(np.average((array -
                                         array_means[:, np.newaxis]) ** 2,
                                        axis=1, weights=weights))
        results['std_%s' % key_name] = array_stds

        if rank:
            results["rank_%s" % key_name] = np.asarray(
                rankdata(-array_means, method='min'), dtype=np.int32)

        return results

    def _process_outputs(self, out, n_splits):
        """return results dict and best dict for given outputs"""

        # if one choose to see train score, "out" will contain train score info
        if self.return_train_score:
            (train_scores, test_scores, test_sample_counts,
             fit_time, score_time, parameters) = zip(*out)
        else:
            (test_scores, test_sample_counts,
             fit_time, score_time, parameters) = zip(*out)

        candidate_params = parameters[::n_splits]
        n_candidates = len(candidate_params)

        results = dict()

        # Computed the (weighted) mean and std for test scores alone
        # NOTE test_sample counts (weights) remain the same for all candidates
        test_sample_counts = np.array(test_sample_counts[:n_splits],
                                      dtype=np.int)

        results = self._store_results(
            results, n_splits, n_candidates, 'test_score',
            test_scores, splits=True, rank=True,
            weights=test_sample_counts if self.iid else None)
        if self.return_train_score:
            results = self._store_results(
                results, n_splits, n_candidates,
                'train_score', train_scores, splits=True)
        results = self._store_results(
            results, n_splits, n_candidates, 'fit_time', fit_time)
        results = self._store_results(
            results, n_splits, n_candidates, 'score_time', score_time)

        best_index = np.flatnonzero(results["rank_test_score"] == 1)[0]

        # Use one MaskedArray and mask all the places where the param is not
        # applicable for that candidate. Use defaultdict as each candidate may
        # not contain all the params
        param_results = defaultdict(partial(MaskedArray,
                                            np.empty(n_candidates,),
                                            mask=True,
                                            dtype=object))
        for cand_i, params in enumerate(candidate_params):
            for name, value in params.items():
                # An all masked empty array gets created for the key
                # `"param_%s" % name` at the first occurence of `name`.
                # Setting the value at an index also unmasks that index
                param_results["param_%s" % name][cand_i] = value

        results.update(param_results)

        # Store a list of param dicts at the key 'params'
        results['params'] = candidate_params

        return results, best_index



from __future__ import print_function
from __future__ import division

import numpy as np
from scipy import sparse
from scipy.optimize import OptimizeResult, nnls
from sklearn.utils import check_X_y
from sklearn.base import RegressorMixin
from sklearn.exceptions import ConvergenceWarning
from sklearn.utils.extmath import safe_sparse_dot
from sklearn.linear_model.base import LinearModel


def _rescale_data(X, y, sample_weight):
    """Rescale data so as to support sample_weight"""
    n_samples = X.shape[0]
    sample_weight = sample_weight * np.ones(n_samples)
    sample_weight = np.sqrt(sample_weight)
    sw_matrix = sparse.dia_matrix((sample_weight, 0),
                                  shape=(n_samples, n_samples))
    X = safe_sparse_dot(sw_matrix, X)
    y = safe_sparse_dot(sw_matrix, y)
    return X, y


class NonNegativeLinearRegression(LinearModel, RegressorMixin):
    """Non-negative least squares linear regression.

    Parameters
    ----------
    fit_intercept : boolean, optional
        whether to calculate the intercept for this model. If set
        to false, no intercept will be used in calculations
        (e.g. data is expected to be already centered).
    normalize : boolean, optional, default False
        If True, the regressors X will be normalized before regression.
        This parameter is ignored when `fit_intercept` is set to False.
        When the regressors are normalized, note that this makes the
        estimated coefficients more robust and almost independent of the
        number of samples. The same property is not valid for standardized
        data. However, if you wish to standardize, please use
        `preprocessing.StandardScaler` before calling `fit` on an estimator
        with `normalize=False`.
    copy_X : boolean, optional, default True
        If True, X will be copied; else, it may be overwritten.

    Attributes
    ----------
    coef_ : array, shape (n_features, )
        Estimated coefficients for the linear regression problem.

    intercept_ : array
        Independent term in the linear model.

    opt_result_ : OptimizeResult
        Result of non-negative least squares optimization
    """

    def __init__(self, fit_intercept=True, normalize=False, copy_X=True):
        self.fit_intercept = fit_intercept
        self.normalize = normalize
        self.copy_X = copy_X

    def fit(self, X, y, sample_weight=None):
        """Fit non-negative linear model.

        Parameters
        ----------
        X : numpy array or sparse matrix of shape [n_samples, n_features]
            Training data
        y : numpy array of shape [n_samples,]
            Target values
        sample_weight : numpy array of shape [n_samples]
            Individual weights for each sample

        Returns
        -------
        self : returns an instance of self.

        """
        X, y = check_X_y(X, y, y_numeric=True, multi_output=False)

        if sample_weight is not None and np.atleast_1d(sample_weight).ndim > 1:
            raise ValueError("Sample weights must be 1D array or scalar")

        X, y, X_offset, y_offset, X_scale = self._preprocess_data(
            X, y, fit_intercept=self.fit_intercept, normalize=self.normalize,
            copy=self.copy_X, sample_weight=sample_weight)

        if sample_weight is not None:
            # Sample weight can be implemented via a simple rescaling.
            X, y = _rescale_data(X, y, sample_weight)

        self.coef_, result = nnls(X, y.squeeze())

        if np.all(self.coef_ == 0):
            raise ConvergenceWarning("All coefficients estimated to be zero in"
                                     " the non-negative least squares fit.")

        self._set_intercept(X_offset, y_offset, X_scale)
        self.opt_result_ = OptimizeResult(success=True, status=0, x=self.coef_,
                                          fun=result)
        return self



from civismlext.stacking import StackedRegressor  # NOQA
from civismlext.stacking import StackedClassifier  # NOQA
from civismlext.nonnegative import NonNegativeLinearRegression  # NOQA
from civismlext.hyperband import HyperbandSearchCV  # NOQA
from civismlext.preprocessing import DataFrameETL  # NOQA
from civismlext._version import __version__  # NOQA



from __future__ import print_function
from __future__ import division

import pytest

from scipy.stats import expon, randint, rankdata
import numpy as np

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.testing import assert_equal, assert_warns
from sklearn.utils.testing import assert_false, assert_true, assert_array_equal
from sklearn.utils import check_random_state
from sklearn.datasets import make_classification
from sklearn.svm import SVC

from ..hyperband import HyperbandSearchCV, hyperband_num_per_run


class DummyCVedEstimator(BaseEstimator, ClassifierMixin):
    def __init__(self, a=1, b=100):
        self.a = a
        self.b = b

    def predict(self, *args, **kwargs):
        pass

    def fit(self, *args, **kwargs):
        self._rng = check_random_state(self.a)
        return self

    def score(self, *args, **kwargs):
        return self._rng.uniform() + self.b


@pytest.mark.parametrize('min_iter', [None, 1, 3])
def test_smoke_hyperband(min_iter):
    seed = 10
    n_splits = 4
    eta = 3
    max_iter = 27
    params = dict(a=randint(low=1, high=100))
    cmax_param = {'b': max_iter}
    if min_iter is not None:
        cmin_param = {'b': min_iter}
    else:
        cmin_param = None
    hyperband_search = HyperbandSearchCV(
        DummyCVedEstimator(),
        cost_parameter_max=cmax_param,
        cost_parameter_min=cmin_param,
        cv=n_splits,
        iid=False,
        return_train_score=False,
        eta=eta,
        param_distributions=params,
        random_state=seed)

    hyperband_search.fit(np.random.normal(size=(100, 3)),
                         np.random.choice(2, size=100, replace=True))

    rng = np.random.RandomState(seed=seed)
    ri = randint(low=1, high=100)

    # b_vals is the geometric sequence of values of b hyperband will test
    if min_iter is None or min_iter == 1:
        b_vals = np.array(
            [1] * 27 + [3] * 9 + [9] * 3 + [27] +
            [3] * 12 + [9] * 4 + [27] +
            [9] * 6 + [27] * 2 +
            [27] * 4)
        a_itrs = [[27, 9, 3, 1], [12, 4, 1], [6, 2], [4]]
    else:
        b_vals = np.array(
            [3] * 9 + [9] * 3 + [27] +
            [9] * 5 + [27] +
            [27] * 3)
        a_itrs = [[9, 3, 1], [5, 1], [3]]

    # now draw the a_vals
    # this will be a list of list of lists of the
    # indices for each score
    inds_to_reorder = []
    loc = 0  # counter used to track the next index
    a_vals = []
    for bstart, nks in enumerate(a_itrs):
        a_vals_orig = ri.rvs(random_state=rng, size=nks[0]).tolist()
        # this first element is a duplicate that is removed below
        _inds_to_reorder = [list(np.arange(loc, loc+nks[0]))]
        for i, nk in enumerate(nks):
            scores = np.array(
                [np.random.RandomState(seed=s).uniform()
                 for s in a_vals_orig]) + np.power(3, i + bstart)
            sinds = np.argsort(scores)[::-1]  # bigger is better
            msk = np.zeros_like(a_vals_orig)
            msk[sinds[0:nk]] = 1
            msk = msk.astype(bool)
            a_vals_orig = [a for i, a in enumerate(a_vals_orig) if msk[i]]
            a_vals += a_vals_orig

            # now add the extra inds
            _inds_to_reorder.append(list(np.arange(loc, loc+nk)))
            loc += nk

        # add the list of lists to the total
        inds_to_reorder.append(_inds_to_reorder[1:])

    a_vals = np.array(a_vals)
    mn_scores = np.array(
        [np.random.RandomState(seed=a).uniform() for a in a_vals]) + b_vals

    # check that we get the same set of scores back
    # a useful test and helps for debugging
    assert (
        set(mn_scores) ==
        set(hyperband_search.cv_results_['mean_test_score'])), (
        "The set of returned scores is wrong!")

    # finally we have to swap the order of the inner and outer loops above
    # to match our implementation of hyperband
    # we have kept the inds corresponding to each iteration, so we just need
    # to add up the lists in the right order
    final_inds = []
    max_inner = max([len(a) for a in inds_to_reorder])
    for rnd in range(max_inner):
        for inds in inds_to_reorder:
            if rnd < len(inds):
                final_inds += inds[rnd]
    # and reorder the outputs
    mn_scores = mn_scores[final_inds]
    a_vals = a_vals[final_inds]
    b_vals = b_vals[final_inds]
    best_index = np.argmax(mn_scores)

    # now make sure it got the right values
    assert hyperband_search.best_index_ == best_index, "Best index is wrong!"
    assert hyperband_search.best_score_ == mn_scores[best_index], (
        "Best score is wrong!")
    assert (hyperband_search.best_params_ ==
            {'a': a_vals[best_index], 'b': b_vals[best_index]}), (
        "Best parameters are wrong!")


def check_cv_results_array_types(cv_results, param_keys, score_keys):
    # Check if the search `cv_results`'s array are of correct types
    assert_true(all(isinstance(cv_results[param], np.ma.MaskedArray)
                    for param in param_keys))
    assert_true(all(cv_results[key].dtype == object for key in param_keys))
    assert_false(any(isinstance(cv_results[key], np.ma.MaskedArray)
                     for key in score_keys))
    assert_true(all(cv_results[key].dtype == np.float64
                    for key in score_keys if not key.startswith('rank')))
    assert_true(cv_results['rank_test_score'].dtype == np.int32)


def check_cv_results_keys(cv_results, param_keys, score_keys, n_cand):
    # Test the search.cv_results_ contains all the required results
    assert_array_equal(sorted(cv_results.keys()),
                       sorted(param_keys + score_keys + ('params',)))
    assert_true(all(cv_results[key].shape == (n_cand,)
                    for key in param_keys + score_keys))


def check_cv_results_grid_scores_consistency(search):
    # TODO Remove for sklearn 0.20
    cv_results = search.cv_results_
    res_scores = np.vstack(list([cv_results["split%d_test_score" % i]
                                 for i in range(search.n_splits_)])).T
    res_means = cv_results["mean_test_score"]
    res_params = cv_results["params"]
    n_cand = len(res_params)
    grid_scores = assert_warns(DeprecationWarning, getattr,
                               search, 'grid_scores_')
    assert_equal(len(grid_scores), n_cand)
    # Check consistency of the structure of grid_scores
    for i in range(n_cand):
        assert_equal(grid_scores[i].parameters, res_params[i])
        assert_array_equal(grid_scores[i].cv_validation_scores,
                           res_scores[i, :])
        assert_array_equal(grid_scores[i].mean_validation_score, res_means[i])


@pytest.mark.parametrize('min_iter', [None, 1, 100])
@pytest.mark.filterwarnings('ignore::sklearn.exceptions.ConvergenceWarning')
def test_hyperband_search_cv_results(min_iter):
    # Make a dataset with a lot of noise to get various kind of prediction
    # errors across CV folds and parameter settings
    X, y = make_classification(n_samples=200, n_features=100, n_informative=3,
                               random_state=0)

    # scipy.stats dists now supports `seed` but sklearn
    # still supports scipy 0.12 which doesn't support the seed.
    # Hence the assertions in the test for hyperband alone should not
    # depend on randomization.
    n_splits = 3
    eta = 3
    max_iter = 1000
    n_cand = hyperband_num_per_run(
        eta,
        max_iter,
        min_iter if min_iter is not None else 1)
    params = dict(C=expon(scale=10), gamma=expon(scale=0.1))
    cmax_param = {'max_iter': max_iter}
    if min_iter is not None:
        cmin_param = {'min_iter': min_iter}
    else:
        cmin_param = None
    hyperband_search = HyperbandSearchCV(
        SVC(),
        cost_parameter_max=cmax_param,
        cost_parameter_min=cmin_param,
        cv=n_splits,
        iid=False,
        eta=eta,
        param_distributions=params)
    hyperband_search.fit(X, y)
    hyperband_search_iid = HyperbandSearchCV(
        SVC(),
        cost_parameter_max=cmax_param,
        cost_parameter_min=cmin_param,
        cv=n_splits,
        iid=True,
        eta=eta,
        param_distributions=params)
    hyperband_search_iid.fit(X, y)

    param_keys = ('param_C', 'param_gamma', 'param_max_iter')
    score_keys = ('mean_test_score', 'mean_train_score',
                  'rank_test_score',
                  'split0_test_score', 'split1_test_score',
                  'split2_test_score',
                  'split0_train_score', 'split1_train_score',
                  'split2_train_score',
                  'std_test_score', 'std_train_score',
                  'mean_fit_time', 'std_fit_time',
                  'mean_score_time', 'std_score_time')

    for search, iid in zip(
            (hyperband_search, hyperband_search_iid), (False, True)):
        assert_equal(iid, search.iid)
        cv_results = search.cv_results_
        # Check results structure
        check_cv_results_array_types(cv_results, param_keys, score_keys)
        check_cv_results_keys(cv_results, param_keys, score_keys, n_cand)
        # For random_search, all the param array vals should be unmasked
        assert_false(any(cv_results['param_C'].mask) or
                     any(cv_results['param_gamma'].mask))
        check_cv_results_grid_scores_consistency(search)


@pytest.mark.parametrize('rank', [False, True])
@pytest.mark.parametrize('splits', [False, True])
@pytest.mark.parametrize('weights', [None, np.array([0.1, 0.4, 0.5])])
@pytest.mark.filterwarnings('ignore::sklearn.exceptions.ConvergenceWarning')
def test_store_results_method(weights, splits, rank):
    rng = np.random.RandomState(seed=10)

    n_candidates = 10
    n_splits = 3
    hyperband_search = HyperbandSearchCV(
        SVC(),
        cost_parameter_max={},
        param_distributions={})

    array = rng.uniform(size=n_candidates * n_splits)
    mns = np.average(array.reshape(10, 3), axis=1, weights=weights)
    stds = np.sqrt(np.average(
        (array.reshape(10, 3) - mns.reshape(-1, 1)) ** 2,
        axis=1,
        weights=weights))
    ranks = np.array(rankdata(-mns, method='min'), dtype=np.int32)
    key_name = 'abc'
    results = {}

    results = hyperband_search._store_results(
        results,
        n_splits,
        n_candidates,
        key_name,
        array,
        weights=weights,
        splits=splits,
        rank=rank)

    if splits:
        for i in range(n_splits):
            assert np.allclose(
                results['split%d_abc' % i],
                array.reshape(10, 3)[:, i]), (
                "Array values for split %d are not right!" % i)

    assert np.allclose(results['mean_abc'], mns), "Means are not correct!"
    assert np.allclose(results['std_abc'], stds), "Stds are not correct!"

    if rank:
        assert np.array_equal(results['rank_abc'], ranks), (
            "Ranks are not correct!")


@pytest.fixture(params=[True, False])
def out(request):
    if request.param:
        # this bool indicates if the train scores are returned as first
        # element of the tuples
        return [(0.1, 0.4, 10, 1, 2, {'a': 1}),
                (0.2, 0.5, 20, 3, 4, {'a': 1}),
                (0.3, 0.6, 11, 5, 6, {'a': 1}),
                (0.7, 1.0, 10, 7, 8, {'a': 2, 'b': 6}),
                (0.8, 1.1, 20, 9, 10, {'a': 2, 'b': 6}),
                (0.9, 1.2, 11, 11, 12, {'a': 2, 'b': 6})]
    else:
        return [(0.4, 10, 1, 2, {'a': 1}),
                (0.5, 20, 3, 4, {'a': 1}),
                (0.6, 11, 5, 6, {'a': 1}),
                (1.0, 10, 7, 8, {'a': 2, 'b': 6}),
                (1.1, 20, 9, 10, {'a': 2, 'b': 6}),
                (1.2, 11, 11, 12, {'a': 2, 'b': 6})]


@pytest.mark.parametrize('iid', [False, True])
@pytest.mark.filterwarnings('ignore::sklearn.exceptions.ConvergenceWarning')
def test_process_outputs_method(out, iid):
    return_train_score = len(out[0]) == 6  # train scores are returned?
    hyperband_search = HyperbandSearchCV(
        SVC(),
        cost_parameter_max={},
        param_distributions={},
        return_train_score=return_train_score,
        iid=iid)

    if iid:
        wts = np.array([10, 20, 11]) / 41
    else:
        wts = None

    score_mns = [
        np.average([0.4, 0.5, 0.6], weights=wts),
        np.average([1.0, 1.1, 1.2], weights=wts)]

    results, best_index = hyperband_search._process_outputs(
        out, n_splits=3)

    assert best_index == 1, (
        "The best candidate index was not computed correctly!")

    assert results['params'] == ({'a': 1}, {'a': 2, 'b': 6})
    assert np.array_equal(
        results['param_a'],
        np.ma.array(np.array([1, 2]), mask=[0, 0], dtype=object)), (
        "Param a results are not correct!")
    # use np.all here since masked array comparison will ignore the masked
    # element - np.array_equal does not for some reason...
    assert np.all(results['param_b'] ==
                  np.ma.array([None, 6], mask=[True, False], dtype=object)), (
            "Param b results are not correct!")

    assert np.allclose(results['split0_test_score'], [0.4, 1.0]), (
        "Split 0 test score is wrong!")
    assert np.allclose(results['split1_test_score'], [0.5, 1.1]), (
        "Split 1 test score is wrong!")
    assert np.allclose(results['split2_test_score'], [0.6, 1.2]), (
        "Split 2 test score is wrong!")
    assert np.allclose(results['mean_test_score'], score_mns), (
        "Mean test score is not correct!")
    assert np.allclose(
        results['std_test_score'],
        [np.sqrt(np.average(
            (np.array([0.4, 0.5, 0.6]) - score_mns[0]) ** 2, weights=wts)),
         np.sqrt(np.average(
            (np.array([1.0, 1.1, 1.2]) - score_mns[1]) ** 2, weights=wts))]), (
        "Std of test score is not correct!")
    assert np.allclose(results['rank_test_score'], [2, 1]), (
        "Test score ranks are wrong!")

    if return_train_score:
        assert np.allclose(results['split0_train_score'], [0.1, 0.7]), (
            "Split 0 train score is wrong!")
        assert np.allclose(results['split1_train_score'], [0.2, 0.8]), (
            "Split 1 train score is wrong!")
        assert np.allclose(results['split2_train_score'], [0.3, 0.9]), (
            "Split 2 train score is wrong!")
        assert np.allclose(results['mean_train_score'], [0.2, 0.8]), (
            "Mean train score is not correct!")
        assert np.allclose(
            results['std_train_score'],
            [np.sqrt(np.average(
                (np.array([0.1, 0.2, 0.3]) - 0.2) ** 2)),
             np.sqrt(np.average(
                (np.array([0.7, 0.8, 0.9]) - 0.8) ** 2))]), (
            "Std of train score is not correct!")

    assert np.allclose(
        results['mean_score_time'],
        [np.mean([2, 4, 6]), np.mean([8, 10, 12])]), (
        "Mean score time is not correct!")

    assert np.allclose(
        results['std_score_time'], [np.std([2, 4, 6]), np.std([8, 10, 12])]), (
        "Std of score time is not correct!")

    assert np.allclose(
        results['mean_fit_time'], [np.mean([1, 3, 5]), np.mean([7, 9, 11])]), (
        "Mean fit time is not correct!")

    assert np.allclose(
        results['std_fit_time'], [np.std([1, 3, 5]), np.std([7, 9, 11])]), (
        "Std of fit time is not correct!")



import os
from setuptools import find_packages, setup


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as _in:
        return _in.read()


__version__ = None
exec(read('civismlext/_version.py'))

setup(version=__version__,
      name="civisml-extensions",
      author="Civis Analytics",
      author_email="opensource@civisanalytics.com",
      url="https://www.civisanalytics.com",
      description="scikit-learn-compatible estimators from Civis Analytics",
      packages=find_packages(),
      install_requires=read('requirements.txt').splitlines(),
      long_description=read('README.rst'),
      include_package_data=True,
      license="BSD-3")



from __future__ import print_function
from __future__ import division

from .. import stacking
from ..stacking import StackedRegressor, StackedClassifier
from ..nonnegative import NonNegativeLinearRegression

import pytest
import numpy as np
from sklearn.tree import DecisionTreeClassifier
from sklearn.model_selection import KFold, train_test_split
from sklearn.metrics import roc_curve, auc, mean_squared_error
from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.datasets import make_regression, make_classification
from numpy.testing import assert_almost_equal, assert_array_equal
import six

if six.PY2:
    import mock
else:
    from unittest import mock


def generate_data(slope=1, magnitude=2, sigma=0.1,
                  rng=np.random.RandomState(7), ntrain=50, ntest=1000,
                  classification=False):
    totsize = ntrain + ntest
    X = rng.normal(scale=5, size=(totsize, 2))

    # Set yint = 0 if even, yint = 1 if odd
    yints = magnitude*(X[:, 0].round() % 2) - magnitude/2

    if classification:
        probs = 1/(1 + np.exp(-(slope*X[:, 0] + 0.1*X[:, 0]*X[:, 1] + yints)))
        y = rng.binomial(1, probs)
    else:
        y = slope*X[:, 0] + yints + rng.normal(scale=sigma, size=totsize)

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=ntest,
                                                        random_state=rng)
    return X_train.reshape(-1, 2), X_test.reshape(-1, 2), y_train, y_test


@pytest.fixture
def clf_test_data():
    rng = np.random.RandomState(7)
    xtrain, xtest, ytrain, ytest = generate_data(slope=0.5,
                                                 classification=True,
                                                 rng=rng)
    return {"x": xtrain, "y": ytrain, "xtest": xtest, "ytest": ytest}


@pytest.fixture
def regression_test_data():
    rng = np.random.RandomState(7)
    xtrain, xtest, ytrain, ytest = generate_data(rng=rng)
    return {"x": xtrain, "y": ytrain, "xtest": xtest, "ytest": ytest}


class NoFit(object):
    """Small class to test parameter dispatching.
    """
    def __init__(self, a=None, b=None):
        self.a = a
        self.b = b


class Transformer(NoFit):
    """Simple transformer-like class
    """
    def fit(self, X, y, **fit_params):
        self.Xfit = X
        self.yfit = y
        self.fit_params = fit_params
        return self

    def get_params(self, deep=False):
        return {'a': self.a, 'b': self.b}

    def set_params(self, **params):
        self.a = params['a']
        return self

    def transform(self, X):
        return X


class BasicEst(NoFit):
    """Simple estimator-like class
    """
    def fit(self, X, y, **fit_params):
        self.Xfit = X
        self.yfit = y
        self.fit_params = fit_params
        return self

    def get_params(self, deep=False):
        return {'a': self.a, 'b': self.b}

    def set_params(self, **params):
        self.a = params['a']
        return self

    def predict(self, X):
        self.Xpred = X
        return np.array([[3, 4],
                         [5, 6],
                         [7, 8]])

    def score(self, X, y):
        pass


class ClfEst(BasicEst):
    """Classifier-like class, with classes_
    """
    def fit(self, X, y, **fit_params):
        super(ClfEst, self).fit(X, y, **fit_params)
        self.classes_ = [0, 1]
        return self

    def predict_proba(self, X):
        self.Xpred = X
        return np.array([[1, 2, 3],
                         [4, 5, 6],
                         [7, 8, 9]])


class FullClfEst(ClfEst):
    """Classifier-like class, with decision_function
    """
    def decision_function(self, X):
        self.Xpred = X
        return np.array([1, 2, 3])


class PassThruClf(ClfEst):
    def predict(self, X):
        raise RuntimeError("you shouldn't be here")

    def predict_proba(self, X):
        return self.Xfit


class PassThruDF(ClfEst):
    def predict(self, X):
        raise RuntimeError("you shouldn't be here")

    def predict_proba(self, X):
        raise RuntimeError("you shouldn't be here")

    def decision_function(self, X):
        return self.Xfit


class PassThruReg(ClfEst):
    def predict(self, X):
        return self.Xfit

    def predict_proba(self, X):
        raise RuntimeError("you shouldn't be here")

    def decision_function(self, X):
        raise RuntimeError("you shouldn't be here")


@pytest.mark.parametrize('SM', [StackedClassifier, StackedRegressor])
def test_init(SM):
    """Construct some StackedModels and check their parameters
    """
    with pytest.raises(TypeError):
        sm = SM()

    # Smoke test with estimator-like object
    basic_est = BasicEst()
    sm = SM([('be', basic_est)], cv=9)  # NOQA


@pytest.mark.parametrize('SM', [StackedClassifier, StackedRegressor])
def test_get_params(SM):
    basic_est = BasicEst()
    meta_est = BasicEst()
    est_list = [('be', basic_est), ('me', meta_est)]

    sm = SM(
        est_list, cv=9, n_jobs=10, pre_dispatch=1, verbose=10)

    # Check parameter getting
    assert sm.get_params(deep=False) == dict(cv=9,
                                             estimator_list=est_list,
                                             n_jobs=10,
                                             pre_dispatch=1,
                                             verbose=10)

    assert sm.get_params(deep=True) == dict(be__a=None, be__b=None,
                                            be=basic_est, me__a=None,
                                            me__b=None, me=meta_est,
                                            **sm.get_params(deep=False))


@pytest.mark.parametrize('SM', [StackedClassifier, StackedRegressor])
def test_get_params_estimator_list_empty(SM):
    est_list = []

    sm = SM(
        est_list, cv=9, n_jobs=10, pre_dispatch=1, verbose=10)

    # Check parameter getting
    param_dict = dict(cv=9,
                      estimator_list=est_list,
                      n_jobs=10,
                      pre_dispatch=1,
                      verbose=10)
    assert sm.get_params(deep=False) == param_dict
    assert sm.get_params(deep=True) == param_dict


@pytest.mark.parametrize('SM', [StackedClassifier, StackedRegressor])
def test_set_params(SM):
    basic_est = BasicEst()
    meta_est = BasicEst()
    est_list = [('be', basic_est), ('me', meta_est)]

    sm = SM(
        est_list, cv=9, n_jobs=10, pre_dispatch=1, verbose=10)

    # Check parameter setting
    sm.set_params(be__a=17)
    assert basic_est.a == 17
    assert sm.get_params(deep=True) == dict(be__a=17, be__b=None,
                                            be=basic_est, me__a=None,
                                            me__b=None, me=meta_est,
                                            **sm.get_params(deep=False))

    # Wrong names should raise exception
    with pytest.raises(ValueError):
        sm.set_params(foo__param=2)


def test_properties_clf():
    """Test that properties get set as expected.
    """
    base_clf = ClfEst(3, 5)
    meta_clf = ClfEst(4, 6)
    sm = StackedClassifier([('be', base_clf), ('me', meta_clf)])

    assert sm.meta_estimator == meta_clf
    assert sm.meta_estimator_name == 'me'
    assert sm.named_base_estimators == {'be': base_clf}

    with pytest.raises(AttributeError):
        sm.classes_
    meta_clf.fit('foo', 'bar')
    assert sm.classes_ == [0, 1]


def test_properties_regression():
    """Test that properties get set as expected.
    """
    base_reg = BasicEst(5, 7)
    meta_reg = BasicEst(6, 8)
    sm = StackedRegressor([('be', base_reg), ('me', meta_reg)])

    assert sm.meta_estimator == meta_reg
    assert sm.meta_estimator_name == 'me'
    assert sm.named_base_estimators == {'be': base_reg}


@pytest.mark.parametrize('n_jobs', [1, 3])
def test_smoke_clf_methods(clf_test_data, n_jobs):
    """Construct, fit, and predict on realistic problem.
    """
    xtrain = clf_test_data['x']
    ytrain = clf_test_data['y']

    rng = np.random.RandomState(17)
    est_list = [('lr', LogisticRegression(C=10**6, random_state=rng)),
                ('rf', RandomForestClassifier(random_state=rng)),
                ('metalr', LogisticRegression(random_state=rng))]
    sm = StackedClassifier(est_list, n_jobs=n_jobs)
    sm.fit(xtrain, ytrain)
    sm.predict(xtrain)
    sm.predict_proba(xtrain)
    sm.predict_log_proba(xtrain)
    sm.decision_function(xtrain)
    sm.score(xtrain, ytrain)
    sm.classes_


@pytest.mark.parametrize('n_jobs', [1, 3])
def test_smoke_multiclass_clf_methods(clf_test_data, n_jobs):
    """Construct, fit, and predict on realistic problem.
    """
    rng = np.random.RandomState(17)
    X, y = make_classification(n_classes=4, n_informative=4, random_state=rng)
    est_list = [('dt', DecisionTreeClassifier(random_state=rng)),
                ('rf', RandomForestClassifier(random_state=rng)),
                ('metarf', RandomForestClassifier(random_state=rng))]
    sm = StackedClassifier(est_list, n_jobs=n_jobs)
    sm.fit(X, y)
    sm.predict(X)
    sm.predict_proba(X)
    sm.predict_log_proba(X)
    sm.score(X, y)
    sm.classes_


@pytest.mark.parametrize('n_jobs', [1, 3])
def test_smoke_regression_methods(regression_test_data, n_jobs):
    """Construct, fit, and predict on realistic problem.
    """
    xtrain = regression_test_data['x']
    ytrain = regression_test_data['y']

    rng = np.random.RandomState(17)
    est_list = [('lr', LinearRegression()),
                ('rf', RandomForestRegressor(random_state=rng)),
                ('nnls', NonNegativeLinearRegression())]
    sm = StackedRegressor(est_list, n_jobs=n_jobs)
    sm.fit(xtrain, ytrain)
    sm.predict(xtrain)
    sm.score(xtrain, ytrain)

    with pytest.raises(AttributeError):
        sm.predict_proba(xtrain)


@pytest.mark.parametrize('n_jobs', [1, 3])
def test_smoke_multiout_regression_methods(n_jobs):
    """Construct, fit, and predict on realistic problem.
    """
    X, y = make_regression(random_state=7, n_samples=100, n_features=10,
                           n_informative=4, n_targets=2)

    rng = np.random.RandomState(17)
    est_list = [('lr', LinearRegression()),
                ('rf', RandomForestRegressor(random_state=rng)),
                ('metalr', LinearRegression())]
    sm = StackedRegressor(est_list, n_jobs=n_jobs)
    sm.fit(X, y)
    sm.predict(X)
    sm.score(X, y)

    with pytest.raises(AttributeError):
        sm.predict_proba(X)


@pytest.mark.parametrize('SM', [StackedClassifier, StackedRegressor])
def test_validate_estimators(SM):
    bad_ests1 = [('be', NoFit()), ('me', ClfEst())]
    bad_ests2 = [('be', ClfEst()), ('me', NoFit())]

    # estimators should have a fit method
    for est_list in [bad_ests1, bad_ests2]:
        sm = SM(est_list)
        errmsg = 'does not have fit method'
        with pytest.raises(TypeError) as err:
            sm._validate_estimators()
        assert errmsg in str(err.value)


def test_validate_clf_estimators():
    bad_clfs1 = [('be', Transformer()), ('me', BasicEst())]
    bad_clfs2 = [('be', BasicEst()), ('me', Transformer())]

    errmsg = "does not have `predict_prob`, `decision_function`, or `predict`"
    for est_list in [bad_clfs1, bad_clfs2]:
        sm = StackedClassifier(est_list)
        with pytest.raises(RuntimeError) as runerr:
            sm._validate_estimators()
        assert errmsg in str(runerr.value)


def test_check_clf_methods():
    est = Transformer()
    with pytest.raises(RuntimeError):
        stacking._check_classifier_methods(est, 'bad_est')

    est = BasicEst()
    stacking._check_classifier_methods(est, 'ok_est')


@pytest.mark.parametrize('SM', [StackedClassifier, StackedRegressor])
def test_validate_estimators_in_fit(SM):
    """This makes sure estimators are validated in the `fit` method itself.
    """
    bad_clfs1 = [('be', ClfEst()), ('me', Transformer())]
    bad_clfs2 = [('be', Transformer()), ('me', ClfEst())]

    # clfs should have either predict_proba, decision_function, or predict
    # method
    if SM == StackedClassifier:
        errmsg = "does not have `predict_prob`, `decision_function`, or `pred"
        for est_list in [bad_clfs1, bad_clfs2]:
            sm = StackedClassifier(est_list)
            with pytest.raises(RuntimeError) as runerr:
                sm.fit([[1]], [1])
            assert errmsg in str(runerr.value)

    bad_ests1 = [('be', NoFit()), ('me', ClfEst())]
    bad_ests2 = [('be', ClfEst()), ('me', NoFit())]

    # estimators should have a fit method
    for est_list in [bad_ests1, bad_ests2]:
        sm = SM(est_list)
        errmsg = 'does not have fit method'
        with pytest.raises(TypeError) as err:
            sm.fit([[1]], [1])
        assert errmsg in str(err.value)


@pytest.mark.parametrize('SM', [StackedClassifier, StackedRegressor])
def test_validate_names(SM):
    # No double underscores in estimator names
    bad_names1 = [('foo__q', BasicEst()), ('bar', BasicEst())]
    # No duplicate names
    bad_names2 = [('foo', BasicEst()), ('foo', BasicEst())]
    exception_strings = ['Estimator names must not contain __',
                         'Names provided are not unique:']

    for bn, estr in zip([bad_names1, bad_names2], exception_strings):
        sm = SM(bn)
        names, estimators = zip(*sm.estimator_list)

        with pytest.raises(ValueError) as valerr:
            sm._validate_names(names)
        assert estr in str(valerr.value)


@pytest.mark.parametrize('SM', [StackedClassifier, StackedRegressor])
def test_validate_names_in_fit(SM):
    """This makes sure names are validated in the `fit` method itself.
    """
    # No double underscores in estimator names
    bad_names1 = [('foo__q', BasicEst()), ('bar', BasicEst())]
    # No duplicate names
    bad_names2 = [('foo', BasicEst()), ('foo', BasicEst())]
    exception_strings = ['Estimator names must not contain __',
                         'Names provided are not unique:']

    for bn, estr in zip([bad_names1, bad_names2], exception_strings):
        sm = SM(bn)

        # Bad names should fail on fit
        with pytest.raises(ValueError) as valerr:
            sm.fit([[1]], [1])
        assert estr in str(valerr.value)

        # Ensure that bad parameter setting after construction fails
        sm = SM([('baz', BasicEst())])
        sm.set_params(**{'estimator_list': bn})

        with pytest.raises(ValueError) as valerr:
            sm.fit([[1]], [1])
        assert estr in str(valerr.value)


@pytest.mark.parametrize('SM', [StackedClassifier, StackedRegressor])
def test_validate_at_least_2_estimators(SM):
    # estimator_list must have at least 2 estimators
    sm = SM([('be', FullClfEst())])

    with pytest.raises(RuntimeError) as runerr:
        sm._validate_estimators()
    assert 'You must have two or more estimators' in str(runerr.value)

    with pytest.raises(RuntimeError) as runerr:
        sm.fit([[1]], [1])
    assert 'You must have two or more estimators' in str(runerr.value)


@pytest.mark.parametrize('SM', [StackedClassifier, StackedRegressor])
def test_extract_fit_params(SM):
    sm = SM([('be', BasicEst()), ('me', BasicEst())])

    assert (sm._extract_fit_params(**{'be__a': 7, 'me__foo': 17}) ==
            {'be': {'a': 7}, 'me': {'foo': 17}})

    # There is no estimator called 'other', so there will be a KeyError
    with pytest.raises(KeyError):
        sm._extract_fit_params(**{'be__a': 7, 'me__foo': 17,
                                  'other__thing': 3})

    # Fit params should have double underscore like: 'est__param'
    with pytest.raises(ValueError):
        sm._extract_fit_params(**{'be__a': 7, 'me__foo': 17, 'me': 3})


@pytest.mark.parametrize('SM', [StackedClassifier, StackedRegressor])
def test_base_est_fit(SM):
    estlist = [('be1', ClfEst()),
               ('be2', ClfEst()),
               ('meta', ClfEst())]
    sm = SM(estlist)
    fp = {'be1__foo': 3, 'be2__bar': 7, 'meta__baz': 'wut'}
    X = 234
    y = 567
    sm._base_est_fit(X, y, **fp)

    # Check base estimator fits
    for est in sm.estimator_list[:-1]:
        assert est[1].Xfit == 234
        assert est[1].yfit == 567
        assert est[1].classes_ == [0, 1]

    # Check fit parameters properly dispatched
    assert sm.estimator_list[0][1].fit_params == {'foo': 3}
    assert sm.estimator_list[1][1].fit_params == {'bar': 7}

    # Check meta-estimator remains unfit
    with pytest.raises(AttributeError):
        sm.meta_estimator.classes_


def test_base_est_predict_clf():
    estlist = [('be1', FullClfEst()),
               ('be2', ClfEst()),
               ('meta', ClfEst())]
    sm = StackedClassifier(estlist)
    Xmeta = sm._base_est_predict(np.array([[-1], [-2], [-3]]))

    # Note that the StackedClassifier should strip off the last column from the
    # ouptut ClfEst's of predict_proba, since the rows of predict_proba always
    # sum to 1.
    assert_array_equal(Xmeta, np.array([[1, 1, 2],
                                        [2, 4, 5],
                                        [3, 7, 8]]))


def test_base_est_predict_reg():
    estlist = [('be1', BasicEst()),
               ('be2', BasicEst()),
               ('meta', BasicEst())]
    sm = StackedRegressor(estlist)
    Xmeta = sm._base_est_predict(np.array([[-1], [-2], [-3]]))

    assert_array_equal(Xmeta, np.array([[3, 4, 3, 4],
                                        [5, 6, 5, 6],
                                        [7, 8, 7, 8]]))


@mock.patch('civismlext.stacking.clone', lambda x: x)
def test_base_est_fit_predict_clf():
    estlist = [('be1', PassThruClf()),
               ('be2', PassThruDF()),
               ('meta', PassThruClf())]
    sc = StackedClassifier(estlist, cv=2)
    fit_params = {'be1__foo': 'f', 'meta__bar': 'b'}
    x = np.arange(16).reshape((8, 2))
    y = np.array([0, 1, 0, 1, 0, 0, 1, 1])

    xmeta, ymeta, fps = sc._base_est_fit_predict(x, y, **fit_params)
    assert_array_equal(ymeta, y)
    # Note that the StackedClassifier should strip off the last column from the
    # ouptut ClfEst's of predict_proba, since the rows of predict_proba always
    # sum to 1.
    assert_array_equal(xmeta, np.array([[8.,  8.,  9.],
                                        [10., 10., 11.],
                                        [12., 12., 13.],
                                        [14., 14., 15.],
                                        [0.,  0.,  1.],
                                        [2.,  2.,  3.],
                                        [4.,  4.,  5.],
                                        [6.,  6.,  7.]]))
    assert sc.estimator_list[0][1].fit_params == {'foo': 'f'}
    assert sc.estimator_list[1][1].fit_params == {}
    assert fps == {'bar': 'b'}

    # Meta estimator should have never been fit
    with pytest.raises(AttributeError):
        sc.meta_estimator.Xfit
    with pytest.raises(AttributeError):
        sc.meta_estimator.fit_params


@mock.patch('civismlext.stacking.clone', lambda x: x)
def test_base_est_fit_predict_multiout_clf():
    estlist = [('be1', PassThruClf()),
               ('be2', PassThruDF()),
               ('meta', PassThruClf())]
    sc = StackedClassifier(estlist, cv=2)
    fit_params = {'be1__foo': 'f', 'meta__bar': 'b'}
    x = np.arange(12).reshape((6, 2))
    y = np.array([0, 1, 2, 1, 0, 2])

    xmeta, ymeta, fps = sc._base_est_fit_predict(x, y, **fit_params)
    assert_array_equal(ymeta, y)
    assert_array_equal(xmeta, np.array([[6.,  6.,  7.],
                                        [8.,  8.,  9.],
                                        [10., 10., 11.],
                                        [0.,  0.,  1.],
                                        [2.,  2.,  3.],
                                        [4.,  4.,  5.]]))
    assert sc.estimator_list[0][1].fit_params == {'foo': 'f'}
    assert sc.estimator_list[1][1].fit_params == {}

    # Meta estimator should have never been fit
    with pytest.raises(AttributeError):
        sc.meta_estimator.Xfit
    with pytest.raises(AttributeError):
        sc.meta_estimator.fit_params


def test_fit_clf():
    estlist = [('be1', PassThruClf()),
               ('be2', PassThruDF()),
               ('meta', PassThruClf())]
    sc = StackedClassifier(estlist, cv=2)
    fit_params = {'be1__foo': 'f', 'meta__bar': 'b'}
    x = np.arange(16).reshape((8, 2))
    y = np.array([0, 1, 0, 1, 0, 0, 1, 1])

    sc.fit(x, y, **fit_params)
    xmeta = sc.meta_estimator.Xfit
    ymeta = sc.meta_estimator.yfit
    assert_array_equal(ymeta, y)
    assert_array_equal(xmeta, np.array([[8.,  8.,  9.],
                                        [10., 10., 11.],
                                        [12., 12., 13.],
                                        [14., 14., 15.],
                                        [0.,  0.,  1.],
                                        [2.,  2.,  3.],
                                        [4.,  4.,  5.],
                                        [6.,  6.,  7.]]))
    for name, est in sc.estimator_list[:-1]:
        assert_array_equal(est.Xfit, x)
        assert_array_equal(est.yfit, y)
    assert sc.estimator_list[0][1].fit_params == {'foo': 'f'}
    assert sc.estimator_list[1][1].fit_params == {}
    assert sc.meta_estimator.fit_params == {'bar': 'b'}


def test_fit_multiclass_clf():
    estlist = [('be1', PassThruClf()),
               ('be2', PassThruDF()),
               ('meta', PassThruClf())]
    sc = StackedClassifier(estlist, cv=2)
    fit_params = {'be1__foo': 'f', 'meta__bar': 'b'}
    x = np.arange(12).reshape((6, 2))
    y = np.array([0, 1, 2, 1, 0, 2])

    sc.fit(x, y, **fit_params)
    xmeta = sc.meta_estimator.Xfit
    ymeta = sc.meta_estimator.yfit
    assert_array_equal(ymeta, y)
    assert_array_equal(xmeta, np.array([[6.,  6.,  7.],
                                        [8.,  8.,  9.],
                                        [10., 10., 11.],
                                        [0.,  0.,  1.],
                                        [2.,  2.,  3.],
                                        [4.,  4.,  5.]]))
    for name, est in sc.estimator_list[:-1]:
        assert_array_equal(est.Xfit, x)
        assert_array_equal(est.yfit, y)
    assert sc.estimator_list[0][1].fit_params == {'foo': 'f'}
    assert sc.estimator_list[1][1].fit_params == {}
    assert sc.meta_estimator.fit_params == {'bar': 'b'}


def test_fit_multiout_clf():
    estlist = [('be1', PassThruClf()),
               ('be2', PassThruDF()),
               ('meta', PassThruClf())]
    sc = StackedClassifier(estlist)
    X = np.arange(3).reshape((3, 1))
    y = np.array([[0, 1],
                  [1, 0],
                  [0, 1]])
    with pytest.raises(NotImplementedError):
        sc.fit(X, y)


@mock.patch('civismlext.stacking.clone', lambda x: x)
def test_base_est_fit_predict_regression():
    estlist = [('be1', PassThruReg()),
               ('be2', PassThruReg()),
               ('meta', PassThruReg())]
    sr = StackedRegressor(estlist, cv=2)
    fit_params = {'be1__foo': 'f', 'meta__bar': 'b'}
    x = np.arange(12).reshape((6, 2))
    y = np.arange(6)

    xmeta, ymeta, fps = sr._base_est_fit_predict(x, y, **fit_params)
    assert_array_equal(ymeta, y)
    assert_array_equal(xmeta, np.array([[6.,  7.,  6.,  7.],
                                        [8.,  9.,  8.,  9.],
                                        [10., 11., 10., 11.],
                                        [0.,  1.,  0.,  1.],
                                        [2.,  3.,  2.,  3.],
                                        [4.,  5.,  4.,  5.]]))
    assert sr.estimator_list[0][1].fit_params == {'foo': 'f'}
    assert sr.estimator_list[1][1].fit_params == {}

    # Meta estimator should have never been fit
    with pytest.raises(AttributeError):
        sr.meta_estimator.Xfit
    with pytest.raises(AttributeError):
        sr.meta_estimator.fit_params


def test_fit_regression():
    estlist = [('be1', PassThruReg()),
               ('be2', PassThruReg()),
               ('meta', PassThruReg())]
    sr = StackedRegressor(estlist, cv=2)
    fit_params = {'be1__foo': 'f', 'meta__bar': 'b'}
    x = np.arange(12).reshape((6, 2))
    y = np.arange(6)

    sr.fit(x, y, **fit_params)

    xmeta = sr.meta_estimator.Xfit
    ymeta = sr.meta_estimator.yfit
    assert_array_equal(ymeta, y)
    assert_array_equal(xmeta, np.array([[6.,  7.,  6.,  7.],
                                        [8.,  9.,  8.,  9.],
                                        [10., 11., 10., 11.],
                                        [0.,  1.,  0.,  1.],
                                        [2.,  3.,  2.,  3.],
                                        [4.,  5.,  4.,  5.]]))
    assert sr.estimator_list[0][1].fit_params == {'foo': 'f'}
    assert sr.estimator_list[1][1].fit_params == {}
    for name, est in sr.estimator_list[:-1]:
        assert_array_equal(est.Xfit, x)
        assert_array_equal(est.yfit, y)
    assert sr.estimator_list[0][1].fit_params == {'foo': 'f'}
    assert sr.estimator_list[1][1].fit_params == {}
    assert sr.meta_estimator.fit_params == {'bar': 'b'}


def test_cv_shuffle_indices():
    """Make sure xmeta and ymeta retain the correct order, even when the CV
    generator is shuffling. This is checking for the bug reported in issue #16.
    """
    estlist = [('be1', PassThruReg()),
               ('be2', PassThruReg()),
               ('meta', PassThruReg())]
    sr = StackedRegressor(estlist, cv=KFold(n_splits=2, shuffle=True))
    x = np.arange(6)
    y = np.arange(6)

    # Suppose the train indices of a 2-fold CV are:
    #  [a, b, c]  and  [d, e, f]
    # Then the test indices are:
    #  [d, e, f]  and  [a, b, c]
    # Since xmeta is just a pass-through of x[train] (horizontally stacked
    # twice, due to the two base estimators) and ymeta is a pass-through of
    # y[test], and since x = y, we should expect that:
    #  xmeta[inds, 0] == ymeta
    #
    inds = np.array([3, 4, 5, 0, 1, 2])

    xmeta, ymeta, _ = sr._base_est_fit_predict(x, y)
    np.testing.assert_equal(xmeta[inds, 0], ymeta)


@mock.patch('civismlext.stacking.clone', lambda x: x)
def test_base_est_fit_predict_multiout_regression():
    estlist = [('be1', PassThruReg()),
               ('be2', PassThruReg()),
               ('meta', PassThruReg())]
    sr = StackedRegressor(estlist, cv=2)
    fit_params = {'be1__foo': 'f', 'meta__bar': 'b'}
    x = np.arange(12).reshape((6, 2))
    y = np.array([[0.1, 0.2], [0.3, 0.4], [0.5, 0.6], [0.7, 0.8],
                  [0.9, 1.0], [1.1, 1.2]])

    xmeta, ymeta, fps = sr._base_est_fit_predict(x, y, **fit_params)
    assert_array_equal(ymeta, y)
    assert_array_equal(xmeta, np.array([[6.,  7.,  6.,  7.],
                                        [8.,  9.,  8.,  9.],
                                        [10., 11., 10., 11.],
                                        [0.,  1.,  0.,  1.],
                                        [2.,  3.,  2.,  3.],
                                        [4.,  5.,  4.,  5.]]))
    assert sr.estimator_list[0][1].fit_params == {'foo': 'f'}
    assert sr.estimator_list[1][1].fit_params == {}

    # Meta estimator should have never been fit
    with pytest.raises(AttributeError):
        sr.meta_estimator.Xfit
    with pytest.raises(AttributeError):
        sr.meta_estimator.fit_params


def test_fit_multiout_regression():
    estlist = [('be1', PassThruReg()),
               ('be2', PassThruReg()),
               ('meta', PassThruReg())]
    sr = StackedRegressor(estlist, cv=2)
    fit_params = {'be1__foo': 'f', 'meta__bar': 'b'}
    x = np.arange(12).reshape((6, 2))
    y = np.array([[0.1, 0.2], [0.3, 0.4], [0.5, 0.6], [0.7, 0.8],
                  [0.9, 1.0], [1.1, 1.2]])

    sr.fit(x, y, **fit_params)

    xmeta = sr.meta_estimator.Xfit
    ymeta = sr.meta_estimator.yfit
    assert_array_equal(ymeta, y)
    assert_array_equal(xmeta, np.array([[6.,  7.,  6.,  7.],
                                        [8.,  9.,  8.,  9.],
                                        [10., 11., 10., 11.],
                                        [0.,  1.,  0.,  1.],
                                        [2.,  3.,  2.,  3.],
                                        [4.,  5.,  4.,  5.]]))
    assert sr.estimator_list[0][1].fit_params == {'foo': 'f'}
    assert sr.estimator_list[1][1].fit_params == {}
    for name, est in sr.estimator_list[:-1]:
        assert_array_equal(est.Xfit, x)
        assert_array_equal(est.yfit, y)
    assert sr.meta_estimator.fit_params == {'bar': 'b'}


def test_fit_pred_simple_regression():
    X = np.arange(6).reshape(-1, 1)
    y = np.arange(1, 7)

    slr = StackedRegressor([('lr1', LinearRegression()),
                            ('lr2', LinearRegression()),
                            ('metalr', NonNegativeLinearRegression())])

    slr.fit(X, y)
    # check that base coefs and ints = 1
    for est in slr.estimator_list[:-1]:
        assert_almost_equal(est[1].coef_, np.array([1]))
        assert_almost_equal(est[1].intercept_, 1)
    # check that sum of meta coefs = 1
    assert_almost_equal(slr.meta_estimator.coef_.sum(), 1)
    assert_almost_equal(slr.meta_estimator.intercept_, 0)


def test_fit_pred_simple_clf():
    X = np.arange(6).reshape(-1, 1)
    X = np.vstack((X, X))
    y = (X > 2).astype(int).ravel()

    sc = StackedClassifier([('dt1', DecisionTreeClassifier()),
                            ('dt2', DecisionTreeClassifier()),
                            ('metadt', DecisionTreeClassifier())])

    sc.fit(X, y)
    # All events below 2.5 should have prob=0; above prob=1
    for basetree in sc.estimator_list[:-1]:
        assert_almost_equal(basetree[1].predict_proba(np.array([[2.4]])),
                            np.array([[1, 0]]))
        assert_almost_equal(basetree[1].predict_proba(np.array([[2.6]])),
                            np.array([[0, 1]]))
    assert_almost_equal(sc.predict_proba(np.array([[2.4]])),
                        np.array([[1, 0]]))
    assert_almost_equal(sc.predict_proba(np.array([[2.6]])),
                        np.array([[0, 1]]))


def test_fit_params_regression(regression_test_data):
    xtrain = regression_test_data['x']
    ytrain = regression_test_data['y']
    sample_weights = [1./len(ytrain)] * len(ytrain)
    fit_params = {'rf__sample_weight': sample_weights}
    sr = StackedRegressor([('rf', RandomForestRegressor(random_state=7)),
                           ('rf2', RandomForestRegressor()),
                           ('meta', NonNegativeLinearRegression())])
    Xmeta, ymeta, _ = sr._base_est_fit_predict(
        xtrain, ytrain, **fit_params)
    assert Xmeta.shape == xtrain.shape


def test_fit_params_clf(clf_test_data):
    xtrain = clf_test_data['x']
    ytrain = clf_test_data['y']
    sample_weights = [1./len(ytrain)] * len(ytrain)
    fit_params = {'rf__sample_weight': sample_weights}
    sr = StackedClassifier([('rf', RandomForestClassifier(random_state=7)),
                           ('lr', LogisticRegression()),
                           ('meta', LogisticRegression())])
    Xmeta, ymeta, _ = sr._base_est_fit_predict(
        xtrain, ytrain, **fit_params)
    assert Xmeta.shape == xtrain.shape


def fit_predict_measure_reg(model, xtrain, ytrain, xtest, ytest):
    model.fit(xtrain, ytrain)
    ypred = model.predict(xtest)
    return mean_squared_error(ytest, ypred)


@pytest.mark.parametrize('n_jobs', [1, 3])
def test_integration_regression(regression_test_data, n_jobs):
    """Construct, fit, and predict on realistic problem. Compare goodness of
    fit of stacked model vs. individual base estimators.
    """
    xtrain = regression_test_data['x']
    ytrain = regression_test_data['y']
    xtest = regression_test_data['xtest']
    ytest = regression_test_data['ytest']

    sr = StackedRegressor([('rf', RandomForestRegressor(random_state=7)),
                           ('lr', LinearRegression()),
                           ('metalr', NonNegativeLinearRegression())],
                          n_jobs=n_jobs)
    rf = RandomForestRegressor(random_state=7)
    lr = LinearRegression()
    sr_mse = fit_predict_measure_reg(sr, xtrain, ytrain, xtest, ytest)
    rf_mse = fit_predict_measure_reg(rf, xtrain, ytrain, xtest, ytest)
    lr_mse = fit_predict_measure_reg(lr, xtrain, ytrain, xtest, ytest)

    # Stacked regressor should perform better than its base estimators on this
    # data.
    assert sr_mse < rf_mse
    assert sr_mse < lr_mse
    assert sr_mse < 1.5    # Sanity check


def fit_predict_measure_clf(model, xtrain, ytrain, xtest, ytest):
    model.fit(xtrain, ytrain)
    ypred = model.predict_proba(xtest)
    fpr, tpr, _ = roc_curve(ytest, ypred[:, 1])
    return auc(fpr, tpr)


@pytest.mark.parametrize('n_jobs', [1, 3])
def test_integration_clf(clf_test_data, n_jobs):
    """Construct, fit, and predict on realistic problem. Compare goodness of
    fit of stacked model vs. individual base estimators.
    """
    xtrain = clf_test_data['x']
    ytrain = clf_test_data['y']
    xtest = clf_test_data['xtest']
    ytest = clf_test_data['ytest']

    sc = StackedClassifier([('rf', RandomForestClassifier()),
                            ('lr', LogisticRegression()),
                            ('metalr', LogisticRegression())],
                           n_jobs=n_jobs)
    sc.set_params(rf__random_state=7, rf__n_estimators=20,
                  lr__random_state=8, metalr__random_state=9,
                  lr__C=10**7, metalr__C=10**7)
    lr = LogisticRegression(C=10**7, random_state=8)
    rf = RandomForestClassifier(n_estimators=20, random_state=7)

    sc_auc = fit_predict_measure_clf(sc, xtrain, ytrain, xtest, ytest)
    lr_auc = fit_predict_measure_clf(lr, xtrain, ytrain, xtest, ytest)
    rf_auc = fit_predict_measure_clf(rf, xtrain, ytrain, xtest, ytest)

    # Sanity check the AUCs of the base estimators
    assert lr_auc > 0.6
    assert rf_auc > 0.6
    # Stacked classifier should perform better than its base estimators on this
    # data.
    assert sc_auc > lr_auc
    assert sc_auc > rf_auc



from __future__ import print_function
from __future__ import division

import numpy as np
from abc import ABCMeta, abstractmethod
import six
from sklearn.base import BaseEstimator, clone
from sklearn.utils.metaestimators import if_delegate_has_method
from sklearn.model_selection import check_cv
from sklearn.utils import tosequence, check_X_y
from sklearn.externals.joblib import Parallel, delayed
from sklearn.model_selection._validation import _index_param_value


def _fit_est(est, X, y, **fit_params):
    return est.fit(X, y, **fit_params)


def _reshape_2d_long(arr):
    # Reshape output so it's always 2-d and long
    if arr.ndim < 2:
        arr = arr.reshape(-1, 1)
    return arr


def _regressor_predict(est, X):
    """standardized predictions for regression"""
    return _reshape_2d_long(est.predict(X))


def _regressor_fit_predict(est, Xtrn, ytrn, Xtst, **fit_params):
    """function for doing fit and predict for regressors"""
    est.fit(Xtrn, ytrn, **fit_params)
    return _regressor_predict(est, Xtst)


def _classifier_predict(est, X):
    # Note: this prefers a decision_function to predict_proba
    # when both are present (e.g. logistic regression), per the
    # convention in CalibratedClassifierCV.
    if hasattr(est, "decision_function"):
        ypred = est.decision_function(X)
    elif hasattr(est, "predict_proba"):
        # predict_proba rows always sum to 1, so drop last col
        ypred = est.predict_proba(X)[:, :-1]
    elif hasattr(est, "predict"):
        # you may want to allow predict for pass-through estimators
        ypred = est.predict(X)
    else:
        raise RuntimeError("Estimator without a `decision_function`, "
                           "`predict_proba`, or `predict` method supplied to a"
                           " StackedClassifier.")

    return _reshape_2d_long(ypred)


def _classifier_fit_predict(est, Xtrn, ytrn, Xtst, **fit_params):
    """function for doing fit and predict for classifiers"""
    est.fit(Xtrn, ytrn, **fit_params)
    return _classifier_predict(est, Xtst)


@six.add_metaclass(ABCMeta)
class BaseStackedModel(BaseEstimator):
    """Abstract base class for StackedClassifier and StackedRegressor. It is
    loosely based on sklearn.pipeline.Pipeline.
    """
    def __init__(self,
                 estimator_list,
                 cv=3,
                 n_jobs=1,
                 pre_dispatch='2*n_jobs',
                 verbose=0):
        self.estimator_list = tosequence(estimator_list)
        self.cv = cv
        self.n_jobs = n_jobs
        self.pre_dispatch = pre_dispatch
        self.verbose = verbose

    def get_params(self, deep=True):
        """Get parameters for this estimator.

        Parameters
        ----------
        deep: boolean, optional (default: True)
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        out : mapping of string to any
            Parameter names mapped to their values.
        """
        # If not deep, just get the params of the BaseStackedModel object
        out = super(BaseStackedModel, self).get_params(deep=False)
        if not deep:
            return out
        # If deep, extract parameters from estimators too
        est_list = getattr(self, 'estimator_list')
        # If est_list is an empty list, don't do anything else
        if len(est_list) > 0:
            out.update(self.named_base_estimators.copy())
            out.update({self.meta_estimator_name: self.meta_estimator}.copy())
            for name, estimator in est_list:
                for key, value in estimator.get_params(deep=True).items():
                    out['%s__%s' % (name, key)] = value
        return out

    def set_params(self, **params):
        """Set the parameters of this estimator.

        Valid parameter keys can be listed with ``get_params()``.

        Returns
        -------
        self
        """
        # Ensure strict ordering of parameter setting:
        # 1. All estimators
        if 'estimator_list' in params:
            setattr(self, 'estimator_list', params.pop('estimator_list'))
        # 2. Estimator replacement
        est_names, _ = zip(*getattr(self, 'estimator_list'))
        for name in list(params.keys()):
            if '__' not in name and name in est_names:
                self._replace_est('estimator_list', name, params.pop(name))
        # 3. Estimator parameters and other initilisation arguments
        super(BaseStackedModel, self).set_params(**params)

        return self

    def fit(self, X, y, **fit_params):
        """Fit the model

        Fit the base estimators on CV folds, then use their prediction on the
        validation folds to train the meta-estimator. Then re-fit base
        estimators on full training set.

        Parameters
        ----------
        X : np.ndarray, list of numbers
            Training data.
        y : np.ndarray, list of numbers
            Training targets.
        **fit_params : dict of {string, object}
            Parameters passed to the ``fit`` method of each estimator, where
            each parameter name is prefixed such that parameter ``p`` for
            estimator ``s`` has key ``s__p``.

        Returns
        -------
        self : BaseStackedModel
            This estimator
        """
        self._validate_estimators()
        X, y = check_X_y(X, y, multi_output=True)

        # Fit base estimators on CV training folds, produce features for
        # meta-estimator from predictions on CV test folds.
        Xmeta, ymeta, meta_params = self._base_est_fit_predict(X, y,
                                                               **fit_params)
        # Fit meta-estimator on test fold predictions of base estimators.
        self.meta_estimator.fit(Xmeta, ymeta, **meta_params)
        # Now fit base estimators again, this time on full training set
        self._base_est_fit(X, y, **fit_params)

        return self

    # _replace_est copied nearly verbatim from sklearn.pipeline._BasePipeline
    # v0.18.1 "_replace_step" method.
    def _replace_est(self, ests_attr, name, new_val):
        # assumes `name` is a valid est name
        new_ests = getattr(self, ests_attr)[:]
        for i, (est_name, _) in enumerate(new_ests):
            if est_name == name:
                new_ests[i] = (name, new_val)
                break
        setattr(self, ests_attr, new_ests)

    # _validate_names copied nearly verbatim from
    # sklearn.pipeline._BasePipeline v0.18.1
    def _validate_names(self, names):
        if len(set(names)) != len(names):
            raise ValueError('Names provided are not unique: '
                             '{0!r}'.format(list(names)))
        invalid_names = set(names).intersection(self.get_params(deep=False))
        if invalid_names:
            raise ValueError('Estimator names conflict with constructor '
                             'arguments: {0!r}'.format(sorted(invalid_names)))
        invalid_names = [name for name in names if '__' in name]
        if invalid_names:
            raise ValueError('Estimator names must not contain __: got '
                             '{0!r}'.format(invalid_names))

    @abstractmethod
    def _validate_estimators(self):
        pass

    def _extract_fit_params(self, **fit_params):
        """Extract fit parameters for each estimator and store in nested dict
        """
        fit_params_ests = dict((name, {}) for name, est in self.estimator_list)
        for pname, pval in fit_params.items():
            est, param = pname.split('__', 1)
            fit_params_ests[est][param] = pval

        return fit_params_ests

    def _base_est_fit(self, X, y, **fit_params):
        """Fit the base estimators on X and y.
        """
        fit_params_ests = self._extract_fit_params(**fit_params)

        _jobs = []
        for name, est in self.estimator_list[:-1]:
            _jobs.append(delayed(_fit_est)(
                clone(est), X, y, **fit_params_ests[name]))

        _out = Parallel(
            n_jobs=self.n_jobs,
            verbose=self.verbose,
            pre_dispatch=self.pre_dispatch)(_jobs)

        for name, _ in self.estimator_list[:-1]:
            self._replace_est('estimator_list', name, _out.pop(0))

    @abstractmethod
    def _check_cv(self, y):
        pass

    def _base_est_fit_predict(self, X, y, **fit_params):
        """Fit the base estimators on CV training folds, and return their
        out-of-sample predictions on the test folds as features for the
        meta-estimator. Also return the fit_params for the meta-estimator.
        """
        y = y.squeeze()
        # Construct CV iterator
        cv = self._check_cv(y=y)
        # Extract CV indices since we need them twice, and un-seeded CV
        # generators with `shuffle=True` split differently each time.
        train_inds = []
        test_inds = []
        for train, test in cv.split(X, y):
            train_inds.append(train)
            test_inds.append(test)

        fit_params_ests = self._extract_fit_params(**fit_params)
        _fit_predict = self._get_fit_predict_function()

        _jobs = []

        # Loop over CV folds to get out-of-sample predictions, which become the
        # features for the meta-estimator.
        for train, test in zip(train_inds, test_inds):
            for name, est in self.estimator_list[:-1]:
                # adapted from sklearn.model_selection._fit_and_predict
                # Adjust length of sample weights
                fit_params_est_adjusted = dict([
                    (k, _index_param_value(X, v, train))
                    for k, v in fit_params_ests[name].items()])

                # Fit estimator on training set and score out-of-sample
                _jobs.append(delayed(_fit_predict)(
                    clone(est),
                    X[train],
                    y[train],
                    X[test],
                    **fit_params_est_adjusted))

        _out = Parallel(
            n_jobs=self.n_jobs,
            verbose=self.verbose,
            pre_dispatch=self.pre_dispatch)(_jobs)

        # Extract the results from joblib
        Xmeta, ymeta = None, None
        for test in test_inds:
            ybase = np.empty((y[test].shape[0], 0))
            for name, est in self.estimator_list[:-1]:
                # Build design matrix out of out-of-sample predictions
                ybase = np.hstack((ybase, _out.pop(0)))

            # Append the test outputs to what will eventually be the features
            # for the meta-estimator.
            if Xmeta is not None:
                ymeta = np.concatenate((ymeta, y[test]))
                Xmeta = np.vstack((Xmeta, ybase))
            else:
                Xmeta = ybase
                ymeta = y[test]

        return Xmeta, ymeta, fit_params_ests[self.meta_estimator_name]

    def _base_est_predict(self, X):
        """Return base estimator predictions on X as meta-features.
        """
        Xmeta = np.empty((X.shape[0], 0))
        for name, est in self.estimator_list[:-1]:
            Xmeta = np.hstack((Xmeta, self._est_predict(est, X)))

        return Xmeta

    @abstractmethod
    def _est_predict(self, est, X):
        """This function ensures that the relevant prediction function is
        consistently called for a given base estimator.
        """
        pass

    @abstractmethod
    def _get_fit_predict_function(self):
        """return the function to be used for the fit and out-of-sample
        predictions"""
        pass

    @if_delegate_has_method(delegate='meta_estimator')
    def predict(self, X):
        """Run predictions through base estimators, then predict with the
        meta-estimator on the output of the base estimators.

        Parameters
        ----------
        X : np.ndarray, list of numbers
            Data to predict on.

        Returns
        -------
        y_pred : array-like
        """
        Xmeta = self._base_est_predict(X)

        return self.meta_estimator.predict(Xmeta)

    @if_delegate_has_method(delegate='meta_estimator')
    def score(self, X, y, **params):
        """Run predictions through base estimators, then score with the
        meta-estimator on the output of the base estimators.

        Parameters
        ----------
        X : np.ndarray, list of numbers
            Data to predict on.

        y : np.ndarray, list of numbers
            Targets for scoring. Must fulfill label requirements for
            meta-estimator.

        params: dict of {string, object}
            Parameters passed to the ``score`` method of the meta-estimator.

        Returns
        -------
        y_score : array-like, shape = [n_samples, n_classes]
        """
        Xmeta = self._base_est_predict(X)

        return self.meta_estimator.score(Xmeta, y, **params)

    @property
    def named_base_estimators(self):
        return dict(self.estimator_list[:-1])

    @property
    def meta_estimator(self):
        return self.estimator_list[-1][-1]

    @property
    def meta_estimator_name(self):
        return self.estimator_list[-1][0]


class StackedClassifier(BaseStackedModel):
    """Builds a stacked classification model from a list of named estimators,
    using the final estimator in the list as the meta-estimator.

    This class takes a list of named estimators, and it fits all but the last
    of these estimators (called the "base estimators") on part of the training
    data passed to ``fit``. The remaining training data is used to create
    out-of-sample predictions from these base estimators. The final named
    estimator, called the meta-estimator, is trained on these out-of-sample
    predictions. This allows the meta-estimator to optimally aggregate the
    predictions of several base estimators, hopefully improving on their
    individual predictive powers.

    It is loosely based on sklearn.pipeline.Pipeline.

    Parameters
    ----------
    estimator_list: list of (str, estimator) tuples
        This contains tuples holding the name and estimator of the desired base
        and meta-estimators. The meta-estimator MUST be the final item in
        the list. The order of the base estimators is irrelevant, as long as
        they occur before the meta-estimator.
    cv : int, cross-validation generator, or iterable, optional (default: 3)
        Determines the cross-validation splitting strategy. Possible inputs for
        cv are:
        - None, to use the default 3-fold cross-validation,
        - integer, to specify the number of folds.
        - An object to be used as a cross-validation generator.
        - An iterable yielding train/test splits.
        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`sklearn.model_selection.StratifiedKFold` is used.
        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.
    n_jobs : int, (default: 1)
        Number of jobs to run in parallel.
    pre_dispatch : int, or string, optional
        Controls the number of jobs that get dispatched during parallel
        execution. Reducing this number can be useful to avoid an
        explosion of memory consumption when more jobs get dispatched
        than CPUs can process. This parameter can be:
            - None, in which case all the jobs are immediately
              created and spawned. Use this for lightweight and
              fast-running jobs, to avoid delays due to on-demand
              spawning of the jobs
            - An int, giving the exact number of total jobs that are
              spawned
            - A string, giving an expression as a function of n_jobs,
              as in '2*n_jobs'
    verbose : integer
        Controls the verbosity: the higher, the more messages. A value
        of 10 gives a moderate level of logging. 50 or more is the most
        amount of logging.

    Attributes
    ----------
    named_base_estimators : dict
        Read-only attribute to access any base estimator by user-given name.
        Keys are estimator names and values are estimators.
    meta_estimator: estimator
        The meta-estimator, provided as a separately accessible property.

    Example
    -------
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> from civismlext.stacking import StackedClassifier
    >>> # Note that the final estimator 'metalr' is the meta-estimator
    >>> estlist = [('rf', RandomForestClassifier()),
    >>>            ('lr', LogisticRegression()),
    >>>            ('metalr', LogisticRegression())]
    >>> mysm = StackedClassifier(estlist)
    >>> # Set some parameters, if you didn't set them at instantiation
    >>> mysm.set_params(rf__random_state=7, lr__random_state=8,
    >>>                 metalr__random_state=9, metalr__C=10**7)
    >>> # Fit
    >>> mysm.fit(Xtrain, ytrain)
    >>> # Predict!
    >>> ypred = mysm.predict_proba(Xtest)
    """
    def __init__(self,
                 estimator_list,
                 cv=3,
                 n_jobs=1,
                 pre_dispatch='2*n_jobs',
                 verbose=0):
        super(StackedClassifier, self).__init__(
            estimator_list, cv, n_jobs, pre_dispatch, verbose)

    def _validate_estimators(self):
        """Validates that the names and methods of the estimators match
        expectiations. Overrides `validate_estimators` in `BaseStackedModel`.
        """
        names, estimators = zip(*self.estimator_list)

        if len(self.estimator_list) < 2:
            raise RuntimeError("You must have two or more estimators to fit a "
                               "StackedClassifier!")

        # validate names
        self._validate_names(names)

        # validate meta-estimator
        if not hasattr(self.meta_estimator, "fit"):
            raise TypeError("Meta-estimator '%s' does not have fit method." %
                            self.meta_estimator_name)
        _check_classifier_methods(self.meta_estimator,
                                  self.meta_estimator_name)

        # validate base estimators
        for name, est in self.estimator_list[:-1]:
            if not hasattr(est, "fit"):
                raise TypeError("Estimator '%s' does not have fit method." %
                                name)
            _check_classifier_methods(est, name)

    def _check_cv(self, y):
        """Overrides base class _check_cv
        """
        # Squeezed target should be 1-dimensional
        if len(y.shape) != 1:
            raise NotImplementedError("StackedClassifier does not currently "
                                      "support multi-column classification "
                                      "problems. If your target is a one-hot "
                                      "encoded multi-class problem, please "
                                      "recast it to a single column.")
        return check_cv(self.cv, y=y, classifier=True)

    def _est_predict(self, est, X):
        """This function ensures that the relevant prediction function is
        consistently called for a given base estimator.
        """
        return _classifier_predict(est, X)

    def _get_fit_predict_function(self):
        """return the function to be used for the fit and out-of-sample
        predictions"""
        return _classifier_fit_predict

    @if_delegate_has_method(delegate='meta_estimator')
    def predict_proba(self, X):
        """Run predictions through base estimators, then predict class
        probabilities with the meta-estimator on the output of the base
        estimators.

        Parameters
        ----------
        X : np.ndarray, list of numbers
            Data to predict on.

        Returns
        -------
        y_proba : array-like, shape = [n_samples, n_classes]
        """
        Xmeta = self._base_est_predict(X)

        return self.meta_estimator.predict_proba(Xmeta)

    @if_delegate_has_method(delegate='meta_estimator')
    def decision_function(self, X):
        """Run predictions through base estimators, then pass the output of the
        base estimators to the meta-estimator's decision_function.

        Parameters
        ----------
        X : np.ndarray, list of numbers
            Data to predict on.

        Returns
        -------
        y_score : array-like, shape = [n_samples, n_classes]
        """
        Xmeta = self._base_est_predict(X)

        return self.meta_estimator.decision_function(Xmeta)

    @if_delegate_has_method(delegate='meta_estimator')
    def predict_log_proba(self, X):
        """Run predictions through base estimators, then predict class log
        probabilities with the meta-estimator on the output of the base
        estimators.

        Parameters
        ----------
        X : np.ndarray, list of numbers
            Data to predict on.

        Returns
        -------
        y_score : array-like, shape = [n_samples, n_classes]
        """
        Xmeta = self._base_est_predict(X)

        return self.meta_estimator.predict_log_proba(Xmeta)

    @property
    def classes_(self):
        return self.meta_estimator.classes_


def _check_classifier_methods(clf, name):
    """Checks whether clf has either a `predict_proba`, `decision_function`, or
    `predict` method. Raises if none of these methods are present.

    Parameters
    ----------
    clf: estimator
        Estimator to check. Expected to be a classifier.

    name: str
        Name of estimator, to be output in case estimator is missing expected
        classification functionality.
    """
    if not (hasattr(clf, "predict_proba") or hasattr(clf, "predict") or
            hasattr(clf, "decision_function")):
        raise RuntimeError("Estimator '%s' does not have `predict_prob`, "
                           "`decision_function`, or `predict` method." % name)


class StackedRegressor(BaseStackedModel):
    """Builds a stacked regression model from a list of named estimators, using
    the final estimator in the list as the meta-estimator.

    This class takes a list of named estimators, and it fits all but the last
    of these estimators (called the "base estimators") on part of the training
    data passed to ``fit``. The remaining training data is used to create
    out-of-sample predictions from these base estimators. The final named
    estimator, called the meta-estimator, is trained on these out-of-sample
    predictions. This allows the meta-estimator to optimally aggregate the
    predictions of several base estimators, hopefully improving on their
    individual predictive powers.

    It is loosely based on sklearn.pipeline.Pipeline.

    Parameters
    ----------
    estimator_list: list of (str, estimator) tuples
        This contains tuples holding the name and estimator of the desired base
        and meta-estimators. The meta-estimator MUST be the final item in
        the list. The order of the base estimators is irrelevant, as long as
        they occur before the meta-estimator.
    cv : int, cross-validation generator, or iterable, optional (default: 3)
        Determines the cross-validation splitting strategy. Possible inputs for
        cv are:
        - None, to use the default 3-fold cross-validation,
        - integer, to specify the number of folds.
        - An object to be used as a cross-validation generator.
        - An iterable yielding train/test splits.
        :class:`sklearn.model_selection.KFold` is used by default for
        regression targets.
        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.
    n_jobs : int, (default: 1)
        Number of jobs to run in parallel.
    pre_dispatch : int, or string, optional
        Controls the number of jobs that get dispatched during parallel
        execution. Reducing this number can be useful to avoid an
        explosion of memory consumption when more jobs get dispatched
        than CPUs can process. This parameter can be:
            - None, in which case all the jobs are immediately
              created and spawned. Use this for lightweight and
              fast-running jobs, to avoid delays due to on-demand
              spawning of the jobs
            - An int, giving the exact number of total jobs that are
              spawned
            - A string, giving an expression as a function of n_jobs,
              as in '2*n_jobs'
    verbose : integer
        Controls the verbosity: the higher, the more messages. A value
        of 10 gives a moderate level of logging. 50 or more is the most
        amount of logging.

    Attributes
    ----------
    named_base_estimators : dict
        Read-only attribute to access any base estimator by user-given name.
        Keys are estimator names and values are estimators.
    meta_estimator: estimator
        The meta-estimator, provided as a separately accessible property.

    Example
    -------
    >>> from sklearn.linear_model import LinearRegression
    >>> from sklearn.ensemble import RandomForestRegressor
    >>> from civismlext.stacking import StackedRegressor
    >>> from civismlext.nonnegative import NonNegativeLinearRegression
    >>> # Note that the final estimator 'meta_nnr' is the meta-estimator
    >>> estlist = [('rf', RandomForestRegressor()),
    >>>            ('lr', LinearRegression()),
    >>>            ('meta_nnr', NonNegativeLinearRegression())]
    >>> mysm = StackedRegressor(estlist)
    >>> # Set some parameters, if you didn't set them at instantiation
    >>> mysm.set_params(rf__random_state=7)
    >>> # Fit
    >>> mysm.fit(Xtrain, ytrain)
    >>> # Predict!
    >>> ypred = mysm.predict(Xtest)
    """
    def __init__(self,
                 estimator_list,
                 cv=3,
                 n_jobs=1,
                 pre_dispatch='2*n_jobs',
                 verbose=0):
        super(StackedRegressor, self).__init__(
            estimator_list, cv, n_jobs, pre_dispatch, verbose)

    def _validate_estimators(self):
        """Validates that the names and methods of the estimators match
        expectations. Overrides `validate_estimators` in `BaseStackedModel`.
        """
        names, estimators = zip(*self.estimator_list)

        if len(self.estimator_list) < 2:
            raise RuntimeError("You must have two or more estimators to fit a "
                               "StackedRegressor!")

        # validate names
        self._validate_names(names)

        # validate meta-estimator
        if not hasattr(self.meta_estimator, "fit"):
            raise TypeError("Meta-estimator '%s' does not have fit method." %
                            self.meta_estimator_name)
        # validate base estimators
        for name, est in self.estimator_list[:-1]:
            if not hasattr(est, "fit"):
                raise TypeError("Estimator '%s' does not have fit method." %
                                name)

    def _check_cv(self, y):
        """Overrides base class _check_cv
        """
        return check_cv(self.cv, y=y, classifier=False)

    def _est_predict(self, est, X):
        """This function ensures that the relevant prediction function is
        consistently called for a given base estimator.
        """
        return _regressor_predict(est, X)

    def _get_fit_predict_function(self):
        """return the function to be used for the fit and out-of-sample
        predictions"""
        return _regressor_fit_predict
